import re
import os
import platform
from glob import glob, escape
import logging
from typing import TYPE_CHECKING, Optional, Union
from collections.abc import Generator, Sequence
from typing_extensions import Literal, TypedDict
import warnings
from concurrent.futures import ThreadPoolExecutor

from numba.typed import List as NumbaList
import numba
import psutil
import numpy as np

from libertem.common.math import prod, make_2D_square, flat_nonzero
from libertem.common import Shape
from libertem.io.dataset.base.file import OffsetsSizes
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, DataSetException, DataSetMeta,
    BasePartition, FileSet, File, make_get_read_ranges,
    Decoder, TilingScheme, default_get_read_ranges,
    DtypeConversionDecoder, IOBackend,
)

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    import numpy.typing as nt


class MIBDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/MIBDatasetParams.schema.json",
        "title": "MIBDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "MIB"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
            "io_backend": {
                "enum": IOBackend.get_supported(),
            },
        },
        "required": ["type", "path"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


def read_hdr_file(path):
    result = {}
    # FIXME: do this open via the io backend!
    with open(path, encoding='utf-8', errors='ignore') as f:
        for line in f:
            if line.startswith("HDR") or line.startswith("End\t"):
                continue
            k, v = line.split("\t", 1)
            k = k.rstrip(':')
            v = v.rstrip("\n")
            result[k] = v
    return result


def is_valid_hdr(path):
    # FIXME: do this open via the io backend!
    with open(path, encoding='utf-8', errors='ignore') as f:
        line = next(f)
        return line.startswith("HDR")


def nav_shape_from_hdr(hdr):
    num_frames, scan_x = (
        int(hdr['Frames in Acquisition (Number)']),
        int(hdr['Frames per Trigger (Number)'])
    )
    nav_shape = (num_frames // scan_x, scan_x)
    return nav_shape


def _pattern(path: str) -> str:
    path, ext = os.path.splitext(path)
    ext = ext.lower()
    if ext == '.mib':
        pattern = "%s*.mib" % (
            re.sub(r'[0-9]+$', '', escape(path))
        )
    elif ext == '.hdr':
        pattern = "%s*.mib" % escape(path)
    else:
        raise DataSetException("unknown extension")
    return pattern


def get_filenames(path: str, disable_glob=False) -> list[str]:
    if disable_glob:
        return [path]
    else:
        return glob(_pattern(path))


def _get_sequence(f: "MIBHeaderReader"):
    return f.fields['sequence_first_image']


def get_image_count_and_sig_shape(
    path: str,
    disable_glob: bool = False,
) -> tuple[int, tuple[int, int]]:
    fns = get_filenames(path, disable_glob=disable_glob)
    headers = MIBDataSet._preread_headers(fns)
    count = 0
    files = []
    for path, fields in headers.items():
        f = MIBHeaderReader(path, fields=fields)
        count += f.fields['num_images']
        files.append(f)
    try:
        first_file = list(sorted(files, key=_get_sequence))[0]
        sig_shape = first_file.fields['image_size']
        return count, sig_shape
    except IndexError:
        raise DataSetException("no files found")


# These encoders takes 2D input/output data - this means we can use
# strides to do slicing and reversing. 2D input data means one output
# row (of bytes) corresponds to one input row (of pixels).


@numba.njit(cache=True)
def encode_u1(inp, out):
    for y in range(out.shape[0]):
        out[y] = inp[y]


@numba.njit(cache=True)
def encode_u2(inp, out):
    for y in range(out.shape[0]):
        row_out = out[y]
        row_in = inp[y]
        for i in range(row_in.shape[0]):
            in_value = row_in[i]
            row_out[i * 2] = (0xFF00 & in_value) >> 8
            row_out[i * 2 + 1] = 0xFF & in_value


@numba.njit(cache=True)
def encode_r1(inp, out):
    for y in range(out.shape[0]):
        row_out = out[y]
        row_in = inp[y]
        for stripe in range(row_out.shape[0] // 8):
            for byte in range(8):
                out_byte = 0
                for bitpos in range(8):
                    value = row_in[64 * stripe + 8 * byte + bitpos] & 1
                    out_byte |= (value << bitpos)
                row_out[(stripe + 1) * 8 - (byte + 1)] = out_byte


@numba.njit(cache=True)
def encode_r6(inp, out):
    for y in range(out.shape[0]):
        row_out = out[y]
        row_in = inp[y]
        for i in range(row_out.shape[0]):
            col = i % 8
            pos = i // 8
            in_pos = (pos + 1) * 8 - col - 1
            row_out[i] = row_in[in_pos]


@numba.njit(cache=True)
def encode_r12(inp, out):
    for y in range(out.shape[0]):
        row_out = out[y]
        row_in = inp[y]
        for i in range(row_in.shape[0]):
            col = i % 4
            pos = i // 4
            in_pos = (pos + 1) * 4 - col - 1
            in_value = row_in[in_pos]
            row_out[i * 2] = (0xFF00 & in_value) >> 8
            row_out[i * 2 + 1] = 0xFF & in_value


@numba.njit(inline='always', cache=True)
def _get_row_start_stop(offset_global, offset_local, stride, row_idx, row_length):
    start = offset_global + offset_local + row_idx * stride
    stop = start + row_length
    return start, stop


@numba.njit(inline='always', cache=True)
def _mib_r24_px_to_bytes(
    bpp, frame_in_file_idx, slice_sig_size, sig_size, sig_origin,
    frame_footer_bytes, frame_header_bytes, file_header_bytes,
    file_idx, read_ranges,
):
    # we are reading a part of a single frame, so we first need to find
    # the offset caused by headers and footers:
    footer_offset = frame_footer_bytes * frame_in_file_idx
    header_offset = frame_header_bytes * (frame_in_file_idx + 1)
    byte_offset = footer_offset + header_offset

    # now let's figure in the current frame index:
    # (go down into the file by full frames; `sig_size`)
    offset = file_header_bytes + byte_offset + frame_in_file_idx * sig_size * bpp // 8

    # offset in px in the current frame:
    sig_origin_bytes = sig_origin * bpp // 8

    start = offset + sig_origin_bytes

    # size of the sig part of the slice:
    sig_size_bytes = slice_sig_size * bpp // 8

    stop = start + sig_size_bytes
    read_ranges.append((file_idx, start, stop))

    # this is the addition for medipix 24bit raw:
    # we read the part from the "second 12bit frame"
    second_frame_offset = sig_size * bpp // 8
    read_ranges.append((file_idx, start + second_frame_offset, stop + second_frame_offset))


mib_r24_get_read_ranges = make_get_read_ranges(px_to_bytes=_mib_r24_px_to_bytes)


@numba.njit(inline="always", cache=True)
def _mib_2x2_tile_block(
    slices_arr, fileset_arr, slice_sig_sizes, sig_origins,
    inner_indices_start, inner_indices_stop, frame_indices, sig_size,
    px_to_bytes, bpp, frame_header_bytes, frame_footer_bytes, file_idxs,
    slice_offset, extra, sig_shape,
):
    """
    Generate read ranges for 2x2 Merlin Quad raw data.

    The arrangement means that reading a contiguous block of data from the file,
    we get data from all four quadrants. The arrangement, and thus resulting
    array, looks like this:

    _________
    | 1 | 2 |
    ---------
    | 3 | 4 |
    ---------

    with the original data layed out like this:

    [4 | 3 | 2 | 1]

    (note that quadrants 3 and 4 are also flipped in x and y direction in the
    resulting array, compared to the original data)

    So if we read one row of raw data, we first get the bottom-most rows from 4
    and 3 first, then the top-most rows from 2 and 1.

    This is similar to how FRMS6 works, and we generate the read ranges in a
    similar way here. In addition to the cut-and-flip from FRMS6, we also have
    the split in x direction in quadrants.
    """
    result = NumbaList()

    # positions in the signal dimensions:
    for slice_idx in range(slices_arr.shape[0]):
        # (offset, size) arrays defining what data to read (in pixels)
        # NOTE: assumes contiguous tiling scheme
        # (i.e. a shape like (1, 1, ..., 1, X1, ..., XN))
        # where X1 is <= the dataset shape at that index, and X2, ..., XN are
        # equal to the dataset shape at that index
        slice_origin = slices_arr[slice_idx][0]
        slice_shape = slices_arr[slice_idx][1]

        read_ranges = NumbaList()

        x_shape = slice_shape[1]
        x_size = x_shape * bpp // 8  # back in bytes
        x_size_half = x_size // 2
        stride = x_size_half * 4

        sig_size_bytes = sig_size * bpp // 8

        y_start = slice_origin[0]
        y_stop = slice_origin[0] + slice_shape[0]

        y_size_half = sig_shape[0] // 2
        y_size = sig_shape[0]

        # inner "depth" loop along the (flat) navigation axis of a tile:
        for i, inner_frame_idx in enumerate(range(inner_indices_start, inner_indices_stop)):
            inner_frame = frame_indices[inner_frame_idx]
            file_idx = file_idxs[i]
            f = fileset_arr[file_idx]
            frame_in_file_idx = inner_frame - f[0]
            file_header_bytes = f[3]

            # we are reading a part of a single frame, so we first need to find
            # the offset caused by headers:
            header_offset = file_header_bytes + frame_header_bytes * (frame_in_file_idx + 1)

            # now let's figure in the current frame index:
            # (go down into the file by full frames; `sig_size`)
            offset = header_offset + frame_in_file_idx * sig_size_bytes

            # in total, we generate depth * 2 * (y_stop - y_start) read ranges per tile
            for y in range(y_start, y_stop):
                if y < y_size_half:

                    # top: no y-flip, no x-flip
                    flip = 0

                    # quadrant 1, left part of the result: we have the three other blocks
                    # in the original data in front of us
                    start, stop = _get_row_start_stop(
                        offset, 3 * x_size_half, stride, y, x_size_half
                    )
                    read_ranges.append((
                        file_idx,
                        start,
                        stop,
                        flip,
                    ))
                    # quadrant 2, right part of the result: we have the two other blocks
                    # in the original data in front of us
                    start, stop = _get_row_start_stop(
                        offset, 2 * x_size_half, stride, y, x_size_half
                    )
                    read_ranges.append((
                        file_idx,
                        start,
                        stop,
                        flip,
                    ))
                else:
                    # bottom: both x and y flip
                    flip = 1
                    y = y_size - y - 1
                    # quadrant 3, left part of the result: we have the one other block
                    # in the original data in front of us
                    start, stop = _get_row_start_stop(
                        offset, 1 * x_size_half, stride, y, x_size_half
                    )
                    read_ranges.append((
                        file_idx,
                        start,
                        stop,
                        flip,
                    ))
                    # quadrant 4, right part of the result: we have the no other blocks
                    # in the original data in front of us
                    start, stop = _get_row_start_stop(offset, 0, stride, y, x_size_half)
                    read_ranges.append((
                        file_idx,
                        start,
                        stop,
                        flip,
                    ))

        # the indices are compressed to the selected frames
        compressed_slice = np.array([
            [slice_offset + inner_indices_start] + [i for i in slice_origin],
            [inner_indices_stop - inner_indices_start] + [i for i in slice_shape],
        ])
        result.append((slice_idx, compressed_slice, read_ranges))

    return result


@numba.njit(inline='always', cache=True, boundscheck=True)
def decode_r1_swap_2x2(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 1bit format: each pixel is actually saved as a single bit. 64 bits
    need to be unpacked together. This is the quad variant.

    Parameters
    ==========
    inp : np.ndarray

    out : np.ndarray
        The output buffer, with the signal dimensions flattened

    idx : int
        The index in the read ranges array

    native_dtype : nt.DtypeLike
        The "native" dtype (format-specific)

    rr : np.ndarray
        A single entry from the read ranges array

    origin : np.ndarray
        The 3D origin of the tile, for example :code:`np.array([2, 0, 0])`

    shape : np.ndarray
        The 3D tileshape, for example :code:`np.array([2, 512, 512])`

    ds_shape : np.ndarray
        The complete ND dataset shape, for example :code:`np.array([32, 32, 512, 512])`
    """
    # in case of 2x2 quad, the index into `out` is not straight `idx`, but
    # we have `2 * shape[1]` read ranges generated for one depth.
    num_rows_tile = shape[1]

    out_3d = out.reshape(out.shape[0], -1, shape[-1])

    # each line in the output array generates two entries in the
    # read ranges. so `(idx // 2) % num_rows_tile` is the correct
    # out_y:
    out_y = (idx // 2) % num_rows_tile
    out_x_start = (idx % 2) * (shape[-1] // 2)

    depth = idx // (num_rows_tile * 2)
    flip = rr[3]

    if flip == 0:
        for stripe in range(inp.shape[0] // 8):
            for byte in range(8):
                inp_byte = inp[(stripe + 1) * 8 - (byte + 1)]
                for bitpos in range(8):
                    out_x = 64 * stripe + 8 * byte + bitpos
                    out_x += out_x_start
                    out_3d[depth, out_y, out_x] = (inp_byte >> bitpos) & 1
    else:
        # flip in x direction:
        x_shape = shape[2]
        x_shape_half = x_shape // 2
        for stripe in range(inp.shape[0] // 8):
            for byte in range(8):
                inp_byte = inp[(stripe + 1) * 8 - (byte + 1)]
                for bitpos in range(8):
                    out_x = out_x_start + x_shape_half - 1 - (64 * stripe + 8 * byte + bitpos)
                    out_3d[depth, out_y, out_x] = (inp_byte >> bitpos) & 1


@numba.njit(inline='always', cache=True, boundscheck=True)
def decode_r6_swap_2x2(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 6bit format: the pixels need to be re-ordered in groups of 8. `inp`
    should have dtype uint8. This is the quad variant.

    Parameters
    ==========
    inp : np.ndarray

    out : np.ndarray
        The output buffer, with the signal dimensions flattened

    idx : int
        The index in the read ranges array

    native_dtype : nt.DtypeLike
        The "native" dtype (format-specific)

    rr : np.ndarray
        A single entry from the read ranges array

    origin : np.ndarray
        The 3D origin of the tile, for example :code:`np.array([2, 0, 0])`

    shape : np.ndarray
        The 3D tileshape, for example :code:`np.array([2, 512, 512])`

    ds_shape : np.ndarray
        The complete ND dataset shape, for example :code:`np.array([32, 32, 512, 512])`
    """
    # in case of 2x2 quad, the index into `out` is not straight `idx`, but
    # we have `2 * shape[1]` read ranges generated for one depth.
    num_rows_tile = shape[1]

    out_3d = out.reshape(out.shape[0], -1, shape[-1])

    # each line in the output array generates two entries in the
    # read ranges. so `(idx // 2) % num_rows_tile` is the correct
    # out_y:
    out_y = (idx // 2) % num_rows_tile
    out_x_start = (idx % 2) * (shape[-1] // 2)

    depth = idx // (num_rows_tile * 2)
    flip = rr[3]

    out_cut = out_3d[depth, out_y, out_x_start:out_x_start + out_3d.shape[2] // 2]

    if flip == 0:
        for i in range(out_cut.shape[0]):
            col = i % 8
            pos = i // 8
            out_pos = (pos + 1) * 8 - col - 1
            out_cut[out_pos] = inp[i]
    else:
        # flip in x direction:
        for i in range(out_cut.shape[0]):
            col = i % 8
            pos = i // 8
            out_pos = (pos + 1) * 8 - col - 1
            out_cut[out_cut.shape[0] - out_pos - 1] = inp[i]


@numba.njit(inline='always', cache=True, boundscheck=True)
def decode_r12_swap_2x2(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 12bit format: the pixels need to be re-ordered in groups of 4. `inp`
    should be an uint8 view on padded big endian 12bit data (">u2").
    This is the quad variant.

    Parameters
    ==========
    inp : np.ndarray

    out : np.ndarray
        The output buffer, with the signal dimensions flattened

    idx : int
        The index in the read ranges array

    native_dtype : nt.DtypeLike
        The "native" dtype (format-specific)

    rr : np.ndarray
        A single entry from the read ranges array

    origin : np.ndarray
        The 3D origin of the tile, for example :code:`np.array([2, 0, 0])`

    shape : np.ndarray
        The 3D tileshape, for example :code:`np.array([2, 512, 512])`

    ds_shape : np.ndarray
        The complete ND dataset shape, for example :code:`np.array([32, 32, 512, 512])`
    """
    # in case of 2x2 quad, the index into `out` is not straight `idx`, but
    # we have `2 * shape[1]` read ranges generated for one depth.
    num_rows_tile = shape[1]

    out_3d = out.reshape(out.shape[0], -1, shape[-1])

    # each line in the output array generates two entries in the
    # read ranges. so `(idx // 2) % num_rows_tile` is the correct
    # out_y:
    out_y = (idx // 2) % num_rows_tile
    out_x_start = (idx % 2) * (shape[-1] // 2)

    depth = idx // (num_rows_tile * 2)
    flip = rr[3]

    out_cut = out_3d[depth, out_y, out_x_start:out_x_start + out_3d.shape[2] // 2]

    if flip == 0:
        for i in range(out_cut.shape[0]):
            col = i % 4
            pos = i // 4
            out_pos = (pos + 1) * 4 - col - 1
            out_cut[out_pos] = (inp[i * 2] << 8) + (inp[i * 2 + 1] << 0)
    else:
        # flip in x direction:
        for i in range(out_cut.shape[0]):
            col = i % 4
            pos = i // 4
            out_pos = (pos + 1) * 4 - col - 1
            out_cut[out_cut.shape[0] - out_pos - 1] = (inp[i * 2] << 8) + (inp[i * 2 + 1] << 0)

    # reference non-quad impl:
    # for i in range(out.shape[1]):
    #     col = i % 4
    #     pos = i // 4
    #     out_pos = (pos + 1) * 4 - col - 1
    #     out[idx, out_pos] = (inp[i * 2] << 8) + (inp[i * 2 + 1] << 0)


mib_2x2_get_read_ranges = make_get_read_ranges(
    read_ranges_tile_block=_mib_2x2_tile_block
)


@numba.njit(inline='always', cache=True)
def decode_r1_swap(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 1bit format: each pixel is actually saved as a single bit. 64 bits
    need to be unpacked together.
    """
    for stripe in range(inp.shape[0] // 8):
        for byte in range(8):
            inp_byte = inp[(stripe + 1) * 8 - (byte + 1)]
            for bitpos in range(8):
                out[idx, 64 * stripe + 8 * byte + bitpos] = (inp_byte >> bitpos) & 1


@numba.njit(inline='always', cache=True)
def decode_r6_swap(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 6bit format: the pixels need to be re-ordered in groups of 8. `inp`
    should have dtype uint8.
    """
    for i in range(out.shape[1]):
        col = i % 8
        pos = i // 8
        out_pos = (pos + 1) * 8 - col - 1
        out[idx, out_pos] = inp[i]


@numba.njit(inline='always', cache=True)
def decode_r12_swap(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 12bit format: the pixels need to be re-ordered in groups of 4. `inp`
    should be an uint8 view on big endian 12bit data (">u2")
    """
    for i in range(out.shape[1]):
        col = i % 4
        pos = i // 4
        out_pos = (pos + 1) * 4 - col - 1
        out[idx, out_pos] = (inp[i * 2] << 8) + (inp[i * 2 + 1] << 0)


@numba.njit(inline='always', cache=True)
def decode_r24_swap(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    RAW 24bit format: a single 24bit consists of two frames that are encoded
    like the RAW 12bit format, the first contains the most significant bits.

    So after a frame header, there are (512, 256) >u2 values, which then
    need to be shuffled like in `decode_r12_swap`.

    This decoder function only works together with mib_r24_get_read_ranges
    which generates twice as many read ranges than normally.
    """
    for i in range(out.shape[1]):
        col = i % 4
        pos = i // 4
        out_pos = (pos + 1) * 4 - col - 1
        out_val = np.uint32((inp[i * 2] << 8) + (inp[i * 2 + 1] << 0))
        if idx % 2 == 0:  # from first frame: most significant bits
            out_val = out_val << 12
        out[idx // 2, out_pos] += out_val


class MIBDecoder(Decoder):
    def __init__(self, header: "HeaderDict"):
        self._kind = header['mib_kind']
        self._dtype = header['dtype']
        self._bit_depth = header['bits_per_pixel']
        self._header = header

    def do_clear(self):
        """
        In case of 24bit raw data, the output buffer needs to be cleared
        before writing, as we can't decode the whole frame in a single call of the decoder.
        The `decode_r24_swap` function needs to be able to add to the output
        buffer, so at the beginning, it needs to be cleared.
        """
        if self._kind == 'r' and self._bit_depth == 24:
            return True
        return False

    def _get_decode_r(self):
        bit_depth = self._bit_depth
        layout = self._header['sensor_layout']
        num_chips = self._header['num_chips']
        if layout == (2, 2) and num_chips == 4:
            if bit_depth == 1:
                return decode_r1_swap_2x2
            elif bit_depth == 6:
                return decode_r6_swap_2x2
            elif bit_depth == 12:
                return decode_r12_swap_2x2
            else:
                raise NotImplementedError(
                    f"bit depth {bit_depth} not implemented for layout {layout}"
                )
        if bit_depth == 1:
            return decode_r1_swap
        elif bit_depth == 6:
            return decode_r6_swap
        elif bit_depth == 12:
            return decode_r12_swap
        elif bit_depth == 24:
            return decode_r24_swap
        else:
            raise ValueError("unknown raw bitdepth")

    def get_decode(self, native_dtype, read_dtype):
        kind = self._kind

        if kind == "u":
            return DtypeConversionDecoder().get_decode(
                native_dtype=native_dtype,
                read_dtype=read_dtype,
            )
        elif kind == "r":
            # FIXME: on big endian systems, these need to be implemented without byteswapping
            return self._get_decode_r()
        else:
            raise RuntimeError("unknown type of MIB file")

    def get_native_dtype(self, inp_native_dtype, read_dtype):
        if self._kind == "u":
            # drop the byteswap from the dtype, if it is there
            return inp_native_dtype.newbyteorder('N')
        else:
            # decode byte-by-byte
            return np.dtype("u1")


MIBKind = Literal['u', 'r']


class HeaderDict(TypedDict):
    header_size_bytes: int
    dtype: "nt.DTypeLike"
    mib_dtype: str
    mib_kind: MIBKind
    bits_per_pixel: int
    image_size: tuple[int, int]
    image_size_bytes: int
    sequence_first_image: int
    filesize: int
    num_images: int
    num_chips: int
    sensor_layout: tuple[int, int]


class MIBHeaderReader:
    def __init__(
        self,
        path: str,
        fields: Optional[HeaderDict] = None,
        sequence_start: int = 0,
    ):
        self.path = path
        if fields is None:
            self._fields = None
        else:
            self._fields = fields
        self._sequence_start = sequence_start

    def __repr__(self) -> str:
        return "<MIBHeaderReader: %s>" % self.path

    @staticmethod
    def _get_np_dtype(dtype: str, bit_depth: int) -> "nt.DTypeLike":
        dtype = dtype.lower()
        num_bits = int(dtype[1:])
        if dtype[0] == "u":
            num_bytes = num_bits // 8
            return np.dtype(">u%d" % num_bytes)
        elif dtype[0] == "r":
            if bit_depth == 1:
                return np.dtype("uint64")
            elif bit_depth == 6:
                return np.dtype("uint8")
            elif bit_depth in (12, 24):  # 24bit raw is two 12bit images after another
                return np.dtype("uint16")
            else:
                raise NotImplementedError("unknown bit depth: %s" % bit_depth)
        else:
            raise NotImplementedError(f"unknown dtype: {dtype}")

    def read_header(self) -> HeaderDict:
        header, filesize = self._read_header_bytes(self.path)
        self._fields = self._parse_header_bytes(header, filesize)
        return self._fields

    @staticmethod
    def _read_header_bytes(path) -> tuple[bytes, int]:
        # FIXME: do this read via the IO backend!
        with open(file=path, mode='rb') as f:
            filesize = os.fstat(f.fileno()).st_size
            return f.read(1024), filesize

    @staticmethod
    def _parse_header_bytes(header: bytes, filesize: int) -> HeaderDict:
        header: str = header.decode(encoding='ascii', errors='ignore')
        parts = header.split(",")
        header_size_bytes = int(parts[2])
        parts = [p
                for p in header[:header_size_bytes].split(",")
                if '\x00' not in p]
        dtype = parts[6].lower()
        mib_kind: Literal['u', 'r']
        # To make mypy Literal check happy...
        if dtype[0] == "r":
            mib_kind = "r"
        elif dtype[0] == "u":
            mib_kind = "u"
        else:
            raise ValueError("unknown kind: %s" % dtype[0])
        image_size = (int(parts[5]), int(parts[4]))
        # FIXME: There can either be threshold values for all chips, or maybe also
        # none. For now, we just make use of the fact that the bit depth is
        # supposed to be the last value.
        bits_per_pixel_raw = int(parts[-1])
        if mib_kind == "u":
            bytes_per_pixel = int(parts[6][1:]) // 8
            image_size_bytes = image_size[0] * image_size[1] * bytes_per_pixel
            num_images = filesize // (
                image_size_bytes + header_size_bytes
            )
        elif mib_kind == "r":
            size_factor = {
                1: 1/8,
                6: 1,
                12: 2,
                24: 4,
            }[bits_per_pixel_raw]
            if bits_per_pixel_raw == 24:
                image_size = (image_size[0], image_size[1] // 2)
            image_size_bytes = int(image_size[0] * image_size[1] * size_factor)
            num_images = filesize // (
                image_size_bytes + header_size_bytes
            )
        num_chips = int(parts[3])

        # 2x2, Nx1, 2x2G, Nx1G -> strip off the 'G' suffix for now:
        # Assumption: layout is width x height

        # We currently only support 1x1 and 2x2 layouts, and support for raw
        # files with 2x2 layout is limited to 1, 6 or 12 bits. Support for 24bit TBD.
        # We don't yet support general Nx1 layouts, or arbitrary "chip select"
        # values (2x2 layout, but only a subset of them is enabled).
        sensor_layout_str = parts[7].replace('G', '').split('x')
        sensor_layout = (
            int(sensor_layout_str[0]),
            int(sensor_layout_str[1]),
        )

        if mib_kind == "r":
            # in raw data, the layout information is not decoded yet - we just get
            # rows from different sensors after one another. The image size is
            # accordingly also different.

            # NOTE: this is still a bit hacky - individual sensors in a layout
            # can be enabled or disabled; because we only support 1x1 and 2x2
            # layouts for now, we just do the "layout based reshaping" only
            # in the 2x2 case:
            if num_chips > 1:
                px_length = image_size[0]
                image_size_orig = image_size
                image_size = (
                    px_length * sensor_layout[1],
                    px_length * sensor_layout[0],
                )
                if prod(image_size_orig) != prod(image_size):
                    raise ValueError(
                        f"invalid sensor layout {sensor_layout} "
                        f"(original image size: {image_size_orig})"
                    )

        return {
            'header_size_bytes': header_size_bytes,
            'dtype': MIBHeaderReader._get_np_dtype(parts[6], bits_per_pixel_raw),
            'mib_dtype': dtype,
            'mib_kind': mib_kind,
            'bits_per_pixel': bits_per_pixel_raw,
            'image_size': image_size,
            'image_size_bytes': image_size_bytes,
            'sequence_first_image': int(parts[1]),
            'filesize': filesize,
            'num_images': num_images,
            'num_chips': num_chips,
            'sensor_layout': sensor_layout,
        }

    @property
    def num_frames(self) -> int:
        return self.fields['num_images']

    @property
    def start_idx(self) -> int:
        return self.fields['sequence_first_image'] - self.sequence_start

    @property
    def sequence_start(self) -> Optional[int]:
        return self._sequence_start

    @sequence_start.setter
    def sequence_start(self, val: int):
        self._sequence_start = val

    @property
    def fields(self) -> HeaderDict:
        if self._fields is None:
            self._fields = self.read_header()
        return self._fields


class MIBFile(File):
    def __init__(self, header, *args, **kwargs):
        self._header = header
        super().__init__(*args, **kwargs)

    def get_array_from_memview(self, mem: memoryview, slicing: OffsetsSizes):
        mem = mem[slicing.file_offset:-slicing.skip_end]
        res = np.frombuffer(mem, dtype="uint8")
        cutoff = self._header['num_images'] * (
            self._header['image_size_bytes'] + self._header['header_size_bytes']
        )
        res = res[:cutoff]
        return res.view(dtype=self._native_dtype).reshape(
            (self.num_frames, -1)
        )[:, slicing.frame_offset:slicing.frame_offset + slicing.frame_size]


class MIBFileSet(FileSet):
    def __init__(
        self,
        header: HeaderDict,
        *args, **kwargs
    ):
        self._header = header
        super().__init__(*args, **kwargs)

    def _clone(self, *args, **kwargs):
        return self.__class__(header=self._header, *args, **kwargs)

    def get_read_ranges(
        self, start_at_frame: int, stop_before_frame: int,
        dtype, tiling_scheme: TilingScheme, sync_offset: int = 0,
        roi: Union[np.ndarray, None] = None,
    ):
        fileset_arr = self.get_as_arr()
        bit_depth = self._header['bits_per_pixel']
        kind = self._header['mib_kind']
        layout = self._header['sensor_layout']
        if kind == "r" and bit_depth == 1:
            # special case for bit-packed 1-bit format:
            bpp = 1
        else:
            bpp = np.dtype(dtype).itemsize * 8
        roi_nonzero = None
        if roi is not None:
            roi_nonzero = flat_nonzero(roi).astype(np.int64)
        kwargs = dict(
            start_at_frame=start_at_frame,
            stop_before_frame=stop_before_frame,
            roi_nonzero=roi_nonzero,
            depth=tiling_scheme.depth,
            slices_arr=tiling_scheme.slices_array,
            fileset_arr=fileset_arr,
            sig_shape=tuple(tiling_scheme.dataset_shape.sig),
            sync_offset=sync_offset,
            bpp=bpp,
            frame_header_bytes=self._frame_header_bytes,
            frame_footer_bytes=self._frame_footer_bytes,
        )
        if layout == (1, 1) or self._header['num_chips'] == 1:
            if kind == "r" and bit_depth in (24,):
                return mib_r24_get_read_ranges(**kwargs)
            else:
                return default_get_read_ranges(**kwargs)
        elif layout == (2, 2):
            if kind == "r":
                return mib_2x2_get_read_ranges(**kwargs)
            else:
                return default_get_read_ranges(**kwargs)
        else:
            raise NotImplementedError(
                f"No support for layout {layout} yet - please contact us!"
            )


class MIBDataSet(DataSet):
    # FIXME include sample file for doctest, see Issue #86
    """
    MIB data sets consist of one or more `.mib` files, and optionally
    a `.hdr` file. The HDR file is used to automatically set the
    `nav_shape` parameter from the fields "Frames per Trigger" and "Frames
    in Acquisition." When loading a MIB data set, you can either specify
    the path to the HDR file, or choose one of the MIB files. The MIB files
    are assumed to follow a naming pattern of some non-numerical prefix,
    and a sequential numerical suffix.

    Note that if you are using a per-pixel or per-scan trigger setup, LiberTEM
    won't be able to deduce the x scanning dimension - in that case, you will
    need to specify the `nav_shape` yourself.

    Currently, we support all integer formats, and most RAW formats. Especially,
    the following configurations are not yet supported for RAW files:

     * Non-2x2 layouts with more than one chip
     * 24bit with more than one chip

    .. versionadded:: 0.9.0
        Support for the raw quad format was added

    Examples
    --------

    >>> # both examples look for files matching /path/to/default*.mib:
    >>> ds1 = ctx.load("mib", path="/path/to/default.hdr")  # doctest: +SKIP
    >>> ds2 = ctx.load("mib", path="/path/to/default64.mib")  # doctest: +SKIP

    Parameters
    ----------
    path: str
        Path to either the .hdr file or one of the .mib files

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Common case: (height, width); but can be any dimensionality

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    disable_glob : bool, default False
        Usually, MIB data sets are stored as a series of .mib files, and we can reliably
        guess the whole set from a single path. If you instead save your data set into
        a single .mib file, and have multiple of these in a single directory with the same prefix
        (for example, a.mib, a1.mib and a2.mib), loading a.mib would include a1.mib and a2.mib
        in the data set. Setting :code:`disable_glob` to :code:`True` will only load the single
        .mib file specified as :code:`path`.
    """
    def __init__(self, path, tileshape=None, scan_size=None, disable_glob=False,
                 nav_shape=None, sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._sig_dims = 2
        self._path = str(path)
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        # handle backwards-compatibility:
        if tileshape is not None:
            warnings.warn(
                "tileshape argument is ignored and will be removed after 0.6.0",
                FutureWarning
            )
        if scan_size is not None:
            warnings.warn(
                "scan_size argument is deprecated. please specify nav_shape instead",
                FutureWarning
            )
            if nav_shape is not None:
                raise ValueError("cannot specify both scan_size and nav_shape")
            self._nav_shape = tuple(scan_size)
        if self._nav_shape is None and not path.lower().endswith(".hdr"):
            raise ValueError(
                    "either nav_shape needs to be passed, or path needs to point to a .hdr file"
                )
        self._filename_cache = None
        self._files_sorted: Optional[Sequence[MIBHeaderReader]] = None
        # ._preread_headers() in _do_initialize() filles self._headers
        self._headers = {}
        self._meta = None
        self._total_filesize = None
        self._sequence_start = None
        self._disable_glob = disable_glob

    def _do_initialize(self):
        filenames = self._filenames()
        self._headers = self._preread_headers(filenames)
        self._files_sorted = list(sorted(self._files(),
                                         key=lambda f: f.fields['sequence_first_image']))

        try:
            first_file = self._files_sorted[0]
        except IndexError:
            raise DataSetException("no files found")

        self._sequence_start = first_file.fields['sequence_first_image']
        # _files_sorted is initially created with _sequence_start = None
        # now that we know the first file we can set the correct value
        for f in self._files_sorted:
            f.sequence_start = self._sequence_start

        if self._nav_shape is None:
            hdr = read_hdr_file(self._path)
            self._nav_shape = nav_shape_from_hdr(hdr)
        if self._sig_shape is None:
            self._sig_shape = first_file.fields['image_size']
        elif int(prod(self._sig_shape)) != int(prod(first_file.fields['image_size'])):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(first_file.fields['image_size']))
            )
        self._sig_dims = len(self._sig_shape)
        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=self._sig_dims)
        dtype = first_file.fields['dtype']
        self._total_filesize = sum(
            f.fields['filesize']
            for f in self._files_sorted
        )
        self._image_count = self._num_images()
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=dtype,
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    def get_diagnostics(self):
        assert self._files_sorted is not None
        first_file = self._files_sorted[0]
        header = first_file.fields
        return [
            {"name": "Bits per pixel",
             "value": str(header['bits_per_pixel'])},
            {"name": "Data kind",
             "value": str(header['mib_kind'])},
            {"name": "Layout",
             "value": str(header['sensor_layout'])},
        ]

    @classmethod
    def get_msg_converter(cls):
        return MIBDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"mib", "hdr"}

    @classmethod
    def detect_params(cls, path, executor):
        pathlow = path.lower()
        if pathlow.endswith(".mib"):
            image_count, sig_shape = executor.run_function(get_image_count_and_sig_shape, path)
            nav_shape = make_2D_square((image_count,))
        elif pathlow.endswith(".hdr") and executor.run_function(is_valid_hdr, path):
            hdr = executor.run_function(read_hdr_file, path)
            image_count, sig_shape = executor.run_function(get_image_count_and_sig_shape, path)
            nav_shape = nav_shape_from_hdr(hdr)
        else:
            return False
        return {
            "parameters": {
                "path": path,
                "nav_shape": nav_shape,
                "sig_shape": sig_shape
            },
            "info": {
                "image_count": image_count,
                "native_sig_shape": sig_shape,
            }
        }

    @staticmethod
    def _preread_headers(filenames):
        # Avoid overhead of creating the Pool on low-file-count datasets
        if len(filenames) > 512:
            # Default ThreadPoolExecutor allocates 5 threads per CPU on the machine
            # In testing this was found to be far too many on Linux, and sometimes
            # a good number on Windows but sometimes too many
            max_workers = 2
            if platform.system() == "Windows":
                num_cores = psutil.cpu_count(logical=False)
                max_workers = max(max_workers, int(num_cores * 0.5))
            with ThreadPoolExecutor(max_workers=max_workers) as p:
                header_and_size = p.map(MIBHeaderReader._read_header_bytes, filenames)
        else:
            header_and_size = tuple(map(MIBHeaderReader._read_header_bytes, filenames))

        res = {}
        for path, (header, filesize) in zip(filenames, header_and_size):
            res[path] = MIBHeaderReader._parse_header_bytes(header, filesize)
        return res

    def _filenames(self):
        if self._filename_cache is not None:
            return self._filename_cache
        fns = get_filenames(self._path, disable_glob=self._disable_glob)
        num_files = len(fns)
        if num_files > 16384:
            warnings.warn(
                f"Saving data in many small files (here: {num_files}) is not efficient, "
                "please increase the \"Images Per File\" parameter when acquiring data. If "
                "this leads to \"Too many open files\" errors, on POSIX systems you can increase "
                "the maximum open files limit using the \"ulimit\" shell command.",
                RuntimeWarning
            )
        self._filename_cache = fns
        return fns

    def _files(self) -> Generator[MIBHeaderReader, None, None]:
        for path, fields in self._headers.items():
            f = MIBHeaderReader(path, fields=fields,
                        sequence_start=self._sequence_start)
            yield f

    def _num_images(self):
        return sum(f.fields['num_images'] for f in self._files_sorted)

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        """
        the shape specified or imprinted by nav_shape from the HDR file
        """
        return self._meta.shape

    def check_valid(self):
        pass

    def get_cache_key(self):
        return {
            "path": self._path,
            # shape is included here because the structure will be invalid if you open
            # the same .mib file with a different nav_shape; should be no issue if you
            # open via the .hdr file
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def get_decoder(self) -> Decoder:
        assert self._files_sorted is not None
        first_file = self._files_sorted[0]
        assert self.meta is not None
        return MIBDecoder(
            header=first_file.fields,
        )

    def _get_fileset(self):
        assert self._sequence_start is not None
        first_file = self._files_sorted[0]
        header_size = first_file.fields['header_size_bytes']
        return MIBFileSet(files=[
            MIBFile(
                path=f.path,
                start_idx=f.start_idx,
                end_idx=f.start_idx + f.num_frames,
                native_dtype=f.fields['dtype'],
                sig_shape=self._meta.shape.sig,
                frame_header=f.fields['header_size_bytes'],
                file_header=0,
                header=f.fields,
            )
            for f in self._files_sorted
        ], header=first_file.fields, frame_header_bytes=header_size)

    def get_base_shape(self, roi: Optional[np.ndarray]) -> tuple[int, ...]:
        # With R-mode files, we are constrained to tile sizes that are a
        # multiple of 64px in the fastest dimension!
        # If we make sure full "x-lines" are taken, we are fine (this is the
        # case by default)
        base_shape = super().get_base_shape(roi)
        assert self.meta is not None and base_shape[-1] == self.meta.shape[-1]
        return base_shape

    def get_partitions(self):
        first_file = self._files_sorted[0]
        fileset = self._get_fileset()
        kind = first_file.fields['mib_kind']
        for part_slice, start, stop in self.get_slices():
            yield MIBPartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                kind=kind,
                bit_depth=first_file.fields['bits_per_pixel'],
                io_backend=self.get_io_backend(),
                decoder=self.get_decoder(),
            )

    def __repr__(self):
        return f"<MIBDataSet of {self.dtype} shape={self.shape}>"


class MIBPartition(BasePartition):
    def __init__(self, kind, bit_depth, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._kind = kind
        self._bit_depth = bit_depth



import numpy as np
from scipy.ndimage import gaussian_filter

from libertem.common.math import prod
from libertem.utils import make_cartesian, make_polar, frame_peaks
import libertem.masks as m


def cbed_frame(
        fy=128, fx=128, zero=None, a=None, b=None, indices=None,
        radius=4, all_equal=False, margin=None):
    if zero is None:
        zero = (fy//2, fx//2)
    zero = np.array(zero)
    if a is None:
        a = (fy//8, 0)
    a = np.array(a)
    if b is None:
        b = make_cartesian(make_polar(a) - (0, np.pi/2))
    b = np.array(b)
    if indices is None:
        indices = np.mgrid[-10:11, -10:11]
    if margin is None:
        margin = radius
    indices, peaks = frame_peaks(fy=fy, fx=fx, zero=zero, a=a, b=b, r=margin, indices=indices)

    data = np.zeros((1, fy, fx), dtype=np.float32)

    dists = np.linalg.norm(peaks - zero, axis=-1)
    max_val = max(dists.max() + 1, len(peaks) + 1)

    for i, p in enumerate(peaks):
        data += m.circular(
            centerX=p[1],
            centerY=p[0],
            imageSizeX=fx,
            imageSizeY=fy,
            radius=radius,
            antialiased=True,
        ) * (1 if all_equal else max(1, max_val - dists[i] + i))

    return (data, indices, peaks)


def hologram_frame(amp, phi,
                   counts=1000.,
                   sampling=5.,
                   visibility=1.,
                   f_angle=30.,
                   gaussian_noise=None,
                   poisson_noise=None):
    """
    Generates holograms using phase and amplitude as an input

    See :ref:`holography app` for detailed application example

    .. versionadded:: 0.3.0

    Notes
    -----
    Theoretical basis for hologram simulations see in:
    Lichte, H., and M. Lehmann. Rep. Prog. Phys. 71 (2008): 016102.
    doi:10.1088/0034-4885/71/1/016102
    :cite:`Lichte2008`

    Parameters
    ----------
    amp, phi: np.ndarray, 2d
        normalized amplitude and phase images of the same shape

    counts: float, default: 1000.
        Number of electron counts in vacuum

    sampling: float, default: 5.
        Hologram fringe sampling (number of pixels per fringe)

    visibility: float, default: 1.
        Hologram fringe visibility (aka fringe contrast)

    f_angle: float, default: 30.
        Angle in degrees of hologram fringes with respect to X-axis

    gaussian_noise: float or int or None, default: None.
        Amount of Gaussian smoothing determined by sigma parameter
        applied to the hologram simulating effect of focus spread or
        PSF of the detector.

    poisson_noise: float or int or None, default: None.
        Amount of Poisson applied to the hologram.

    Returns
    -------
    holo: np.ndarray, 2d
        hologram image
    """
    if not amp.shape == phi.shape:
        raise ValueError('Amplitude and phase should be 2d arrays of the same shape.')
    sy, sx = phi.shape
    x, y = np.meshgrid(np.arange(sx), np.arange(sy))
    f_angle = f_angle / 180. * np.pi

    holo = counts / 2 * (1. + amp ** 2 + 2. * amp * visibility
                         * np.cos(2. * np.pi * y / sampling * np.cos(f_angle)
                                  + 2. * np.pi * x / sampling * np.sin(f_angle)
                                  - phi))

    if poisson_noise:
        if not isinstance(poisson_noise, (float, int)):
            raise ValueError("poisson_noise parameter should be float or int or None.")
        noise_scale = poisson_noise * counts
        holo = noise_scale * np.random.poisson(holo / noise_scale)

    if gaussian_noise:
        if not isinstance(gaussian_noise, (float, int)):
            raise ValueError("gaussian_noise parameter should be float or int or None.")
        holo = gaussian_filter(holo, gaussian_noise)

    return holo


def gradient_data(nav_dims, sig_dims):
    data = np.linspace(
        start=5, stop=30, num=prod(nav_dims) * prod(sig_dims), dtype=np.float32
    )
    return data.reshape(nav_dims + sig_dims)


def exclude_pixels(sig_dims, num_excluded):
    '''
    Generate a list of excluded pixels that
    can be reconstructed faithfully from their neighbors
    in a linear gradient dataset
    '''
    if num_excluded == 0:
        return None
    # Map of pixels that can be reconstructed faithfully from neighbors in a linear gradient
    free_map = np.ones(sig_dims, dtype=bool)

    # Exclude all border pixels
    for dim in range(len(sig_dims)):
        selector = tuple(slice(None) if i != dim else (0, -1) for i in range(len(sig_dims)))
        free_map[selector] = False

    exclude = []

    while len(exclude) < num_excluded:
        exclude_item = tuple(np.random.randint(low=1, high=s-1) for s in sig_dims)
        if free_map[exclude_item]:
            exclude.append(exclude_item)
            knock_out = tuple(slice(e - 1, e + 2) for e in exclude_item)
            # Remove the neighbors of a bad pixel
            # since that can't be reconstructed faithfully from a linear gradient
            free_map[knock_out] = False

    # Transform from list of tuples with length of number of dimensions
    # to array of indices per dimension
    return np.array(exclude).T



import typing
import numpy as np

from libertem.common import Slice
from libertem.common.math import count_nonzero, ndenumerate

if typing.TYPE_CHECKING:
    from libertem.io.dataset.base.partition import Partition


def _roi_to_nd_indices(roi, part_slice: Slice):
    """
    Helper function to calculate indices from roi mask

    Parameters
    ----------

    roi : numpy.ndarray of type bool, matching the navigation shape of the dataset

    part_slice
        Slice indicating what part of the roi to operate on, for example,
        corresponding to a partition.
    """
    roi_slice = roi[part_slice.get(nav_only=True)]
    nav_dims = part_slice.shape.nav.dims
    total = 0
    frames_in_roi = count_nonzero(roi)
    for idx, value in ndenumerate(roi_slice):
        if not value:
            continue
        yield tuple(a + b
                    for a, b in zip(idx, part_slice.origin[:nav_dims]))
        # early exit: we know we don't have more frames in the roi
        total += 1
        if total == frames_in_roi:
            break


def roi_for_partition(roi: np.ndarray, partition: 'Partition') -> np.ndarray:
    return roi.reshape(-1)[partition.slice.get(nav_only=True)]



import pytest
import numpy as np
from numpy.testing import assert_allclose

from libertem.io.dataset.base.decode import decode_swap_2, default_decode
from libertem.io.dataset.mib import (
    encode_r1,
    encode_r6,
    encode_r12,
    encode_u1,
    encode_u2,
    decode_r1_swap,
    decode_r6_swap,
    decode_r12_swap,
)


def encode_roundtrip(encode, decode, bits_per_pixel, shape=(512, 512)):
    max_value = (1 << bits_per_pixel) - 1
    data = np.random.randint(0, max_value + 1, shape)
    encoded = np.zeros(data.size // 8 * bits_per_pixel, dtype=np.uint8)
    encoded = encoded.reshape((shape[0], -1))
    encode(inp=data, out=encoded)
    decoded = np.zeros_like(data).reshape((1, -1))
    decode(
        inp=encoded.reshape((-1,)),
        out=decoded,
        idx=0,
        # the parameters below are not used by the mib decoders:
        native_dtype=np.uint8,
        rr=np.zeros(1, dtype=np.uint64),
        origin=np.zeros(1, dtype=np.uint64),
        shape=np.zeros(1, dtype=np.uint64),
        ds_shape=np.zeros(1, dtype=np.uint64),
    )
    decoded = decoded.reshape(data.shape)
    return data, decoded


@pytest.mark.with_numba
@pytest.mark.parametrize(
    'encode,decode,bits_per_pixel', [
        (encode_r1, decode_r1_swap, 1),
        (encode_r6, decode_r6_swap, 8),
        (encode_r12, decode_r12_swap, 16),
        (encode_u1,  default_decode, 8),
        (encode_u2,  decode_swap_2, 16),
    ],
)
def test_encode_roundtrip(encode, decode, bits_per_pixel):
    data, decoded = encode_roundtrip(encode, decode, bits_per_pixel, shape=(256, 256))
    assert_allclose(data, decoded)



import pytest
from collections.abc import Sequence
import numpy as np
import scipy.sparse as sp
import sparse
from sparseconverter import (
    CUPY, CUPY_BACKENDS, NUMPY, SPARSE_BACKENDS, SPARSE_COO, get_device_class
)

from utils import _naive_mask_apply, _mk_random, set_device_class

from libertem.masks import circular
from libertem.common.sparse import to_dense, to_sparse, is_sparse
from libertem.common import Shape, Slice
from libertem.io.dataset.memory import MemoryDataSet
from libertem.udf.masks import ApplyMasksUDF
from libertem.udf import UDF, UDFMeta


def _run_mask_test_program(lt_ctx, dataset, mask, expected, do_sparse=True):
    dtype = UDF.USE_NATIVE_DTYPE

    analysis_default = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask], dtype=dtype
    )
    if do_sparse:
        analysis_sparse = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=[lambda: to_sparse(mask)], use_sparse=True,
            dtype=dtype
        )
    analysis_dense = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: to_dense(mask)], use_sparse=False,
        dtype=dtype
    )
    results_default = lt_ctx.run(analysis_default)
    if do_sparse:
        results_sparse = lt_ctx.run(analysis_sparse)
    results_dense = lt_ctx.run(analysis_dense)

    analysis_list_of_arr = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: [mask, mask], dtype=dtype
    )
    lt_ctx.run(analysis_list_of_arr)

    assert np.allclose(
        results_default.mask_0.raw_data,
        expected
    )
    if do_sparse:
        assert np.allclose(
            results_sparse.mask_0.raw_data,
            expected
        )
    assert np.allclose(
        results_dense.mask_0.raw_data,
        expected
    )


@pytest.mark.xfail
@pytest.mark.slow
def test_weird_partition_shapes_1_slow(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask = _mk_random(size=(16, 16))
    expected = _naive_mask_apply([mask], data)

    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), partition_shape=(16, 16, 2, 2))

    _run_mask_test_program(lt_ctx, dataset, mask, expected)

    p = next(dataset.get_partitions())
    t = next(p.get_tiles())
    assert tuple(t.tile_slice.shape) == (1, 1, 2, 2)


@pytest.mark.xfail
def test_weird_partition_shapes_1_fast(lt_ctx):
    # XXX MemoryDataSet is now using Partition3D and so on, so we can't create
    # partitions with weird shapes so easily anymore (in this case, partitioned in
    # the signal dimensions). maybe fix this with a custom DataSet impl that simulates this?
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask = _mk_random(size=(16, 16))
    expected = _naive_mask_apply([mask], data)

    dataset = MemoryDataSet(data=data, tileshape=(8, 16, 16), partition_shape=(16, 16, 8, 8))

    _run_mask_test_program(lt_ctx, dataset, mask, expected)

    p = next(dataset.get_partitions())
    t = next(p.get_tiles())
    assert tuple(t.tile_slice.shape) == (1, 8, 8, 8)


def test_normal_partition_shape(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask = _mk_random(size=(16, 16))
    expected = _naive_mask_apply([mask], data)

    dataset = MemoryDataSet(data=data, tileshape=(1, 16, 16), num_partitions=2)

    _run_mask_test_program(lt_ctx, dataset, mask, expected)


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
@pytest.mark.parametrize(
    'dtype', ("<u2", "float32")
)
def test_single_frame_tiles(lt_ctx, backend, dtype):
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random(size=(16, 16, 16, 16), dtype=dtype, array_backend=SPARSE_COO)
        else:
            data = _mk_random(size=(16, 16, 16, 16), dtype=dtype, array_backend=NUMPY)

        mask = _mk_random(size=(16, 16))
        expected = _naive_mask_apply([mask], data)

        dataset = MemoryDataSet(
            data=data,
            tileshape=(1, 16, 16),
            num_partitions=2,
            array_backends=(backend, ) if backend is not None else None,
        )

        _run_mask_test_program(lt_ctx, dataset, mask, expected)


@pytest.mark.slow
def test_subframe_tiles_slow(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask = _mk_random(size=(16, 16))
    expected = _naive_mask_apply([mask], data)

    dataset = MemoryDataSet(data=data, tileshape=(1, 4, 4), num_partitions=2)

    _run_mask_test_program(lt_ctx, dataset, mask, expected)


def test_subframe_tiles_fast(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask = _mk_random(size=(16, 16))
    expected = _naive_mask_apply([mask], data)

    dataset = MemoryDataSet(data=data, tileshape=(8, 4, 4), num_partitions=2)

    _run_mask_test_program(lt_ctx, dataset, mask, expected)


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_mask_uint(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random(size=(16, 16, 16, 16), dtype="<u2", array_backend=SPARSE_COO)
        else:
            data = _mk_random(size=(16, 16, 16, 16), dtype="<u2", array_backend=NUMPY)

        mask = _mk_random(size=(16, 16)).astype("uint16")
        expected = _naive_mask_apply([mask], data)

        dataset = MemoryDataSet(
            data=data,
            tileshape=(4 * 4, 4, 4),
            num_partitions=2,
            array_backends=(backend, ) if backend is not None else None,
        )
        # We skip the sparse test with integer masks on CuPy
        # since cuyx.scipy.sparse doesn't support it
        do_sparse = backend not in CUPY_BACKENDS
        _run_mask_test_program(lt_ctx, dataset, mask, expected, do_sparse=do_sparse)


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_endian(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = np.random.choice(a=0xFFFF, size=(16, 16, 16, 16)).astype(">u2")
        mask = _mk_random(size=(16, 16))
        expected = _naive_mask_apply([mask], data)

        dataset = MemoryDataSet(
            data=data,
            tileshape=(4 * 4, 4, 4),
            num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )

        _run_mask_test_program(lt_ctx, dataset, mask, expected)


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_signed(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = np.random.choice(a=0xFFFF, size=(16, 16, 16, 16)).astype("<i4")
        mask = _mk_random(size=(16, 16))
        expected = _naive_mask_apply([mask], data)

        # NOTE: we allow casting from int32 to float32 here, and may lose some
        # precision in case of data with large dynamic range
        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            check_cast=False,
            array_backends=(backend, ) if backend is not None else None,
        )

        _run_mask_test_program(lt_ctx, dataset, mask, expected)


@pytest.mark.parametrize(
    # More in-depth tests below
    'backend', (None, NUMPY, CUPY)
)
def test_multi_masks(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        mask0 = _mk_random(size=(16, 16))
        mask1 = sp.csr_matrix(_mk_random(size=(16, 16)))
        mask2 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
        expected = _naive_mask_apply([mask0, mask1, mask2], data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None,
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=[lambda: mask0, lambda: mask1, lambda: mask2],
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )
        assert np.allclose(
            results.mask_2.raw_data,
            expected[2],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_multi_mask_stack_dense(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, mask_count=2,
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_multi_mask_stack_sparse(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = sparse.COO.from_numpy(_mk_random(size=(2, 16, 16)))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, mask_count=2,
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_multi_mask_stack_force_sparse(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse=True, mask_count=2
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
@pytest.mark.with_numba  # coverage for rmatmul implementation
def test_multi_mask_stack_force_scipy_sparse(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse='scipy.sparse', mask_count=2
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
@pytest.mark.with_numba  # coverage for rmatmul implementation
def test_multi_mask_stack_force_scipy_sparse_csc(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse='scipy.sparse.csc', mask_count=2
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_multi_mask_stack_force_sparse_pydata(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse='sparse.pydata', mask_count=2
        )
        if backend in CUPY_BACKENDS:
            with pytest.raises(ValueError):
                results = lt_ctx.run(analysis)
        else:
            results = lt_ctx.run(analysis)
            assert np.allclose(
                results.mask_0.raw_data,
                expected[0],
            )
            assert np.allclose(
                results.mask_1.raw_data,
                expected[1],
            )


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(ApplyMasksUDF(lambda x: None).get_backends())
)
def test_multi_mask_stack_force_sparse_pydata_GCXS(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = _mk_random(size=(2, 16, 16))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse='sparse.pydata.GCXS', mask_count=2
        )
        if backend in CUPY_BACKENDS:
            with pytest.raises(ValueError):
                results = lt_ctx.run(analysis)
        else:
            results = lt_ctx.run(analysis)
            assert np.allclose(
                results.mask_0.raw_data,
                expected[0],
            )
            assert np.allclose(
                results.mask_1.raw_data,
                expected[1],
            )


@pytest.mark.parametrize(
    'backend', ['numpy', 'cupy']
)
def test_multi_mask_stack_force_dense(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        masks = sparse.COO.from_numpy(_mk_random(size=(2, 16, 16)))
        expected = _naive_mask_apply(masks, data)

        dataset = MemoryDataSet(
            data=data, tileshape=(4 * 4, 4, 4), num_partitions=2,
            array_backends=(backend, ) if backend is not None else None
        )
        analysis = lt_ctx.create_mask_analysis(
            dataset=dataset, factories=lambda: masks, use_sparse=False, mask_count=2
        )
        results = lt_ctx.run(analysis)

        assert np.allclose(
            results.mask_0.raw_data,
            expected[0],
        )
        assert np.allclose(
            results.mask_1.raw_data,
            expected[1],
        )


def test_multi_mask_autodtype(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    masks = _mk_random(size=(2, 16, 16))
    expected = _naive_mask_apply(masks, data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: masks
    )
    results = lt_ctx.run(analysis)

    assert results.mask_0.raw_data.dtype == np.result_type(np.float32, data.dtype, masks.dtype)

    assert np.allclose(
        results.mask_0.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1.raw_data,
        expected[1],
    )


def test_multi_mask_autodtype_wide(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="int64")
    masks = _mk_random(size=(2, 16, 16))
    expected = _naive_mask_apply(masks, data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: masks
    )
    results = lt_ctx.run(analysis)

    assert results.mask_0.raw_data.dtype == np.result_type(np.float64, data.dtype, masks.dtype)

    assert np.allclose(
        results.mask_0.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1.raw_data,
        expected[1],
    )


def test_multi_mask_autodtype_complex(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="complex64")
    masks = _mk_random(size=(2, 16, 16))
    expected = _naive_mask_apply(masks, data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(dataset=dataset, factories=lambda: masks)
    results = lt_ctx.run(analysis)

    assert results.mask_0_complex.raw_data.dtype.kind == 'c'
    assert results.mask_0_complex.raw_data.dtype == np.complex64

    assert np.allclose(
        results.mask_0_complex.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1_complex.raw_data,
        expected[1],
    )


def test_multi_mask_autodtype_complex_wide(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16))
    masks = _mk_random(size=(2, 16, 16), dtype="complex128")
    expected = _naive_mask_apply(masks, data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: masks
    )
    results = lt_ctx.run(analysis)

    assert results.mask_0_complex.raw_data.dtype.kind == 'c'
    assert results.mask_0_complex.raw_data.dtype == np.complex128

    assert np.allclose(
        results.mask_0_complex.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1_complex.raw_data,
        expected[1],
    )


def test_multi_mask_force_dtype(lt_ctx):
    force_dtype = np.dtype(np.int32)
    data = _mk_random(size=(16, 16, 16, 16), dtype="int16")
    masks = _mk_random(size=(2, 16, 16), dtype="bool")
    expected = _naive_mask_apply(masks.astype(force_dtype), data.astype(force_dtype))

    dataset = MemoryDataSet(
        data=data,
        tileshape=(4 * 4, 4, 4),
        num_partitions=2,
        # disable other input dtypes to not upcast to floats
        # for cupyx.scipy.sparse
        array_backends=(NUMPY, ),
    )
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: masks, dtype=force_dtype
    )
    results = lt_ctx.run(analysis)

    assert results.mask_0.raw_data.dtype.kind == force_dtype.kind
    assert results.mask_0.raw_data.dtype == force_dtype

    assert np.allclose(
        results.mask_0.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1.raw_data,
        expected[1],
    )


def test_avoid_calculating_masks_on_client(hdf5_ds_1, local_cluster_ctx):
    mask = _mk_random(size=(16, 16))
    # We have to start a local cluster so that the masks are
    # computed in a different process
    analysis = local_cluster_ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask], mask_count=1, mask_dtype=np.float32
    )
    udf = analysis.get_udf()
    local_cluster_ctx.run_udf(dataset=hdf5_ds_1, udf=udf)

    assert udf.masks._computed_masks is None


def test_avoid_calculating_masks_on_client_udf(hdf5_ds_1, local_cluster_ctx):
    mask = _mk_random(size=(16, 16))
    # We have to use a real cluster instead of InlineJobExecutor so that the masks are
    # computed in a different process
    analysis = local_cluster_ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask], mask_count=1, mask_dtype=np.float32
    )
    udf = analysis.get_udf()
    local_cluster_ctx.run_udf(udf=udf, dataset=hdf5_ds_1)
    assert udf._mask_container is None


def test_override_mask_dtype(lt_ctx):
    mask_dtype = np.float32
    data = _mk_random(size=(16, 16, 16, 16), dtype=mask_dtype)
    masks = _mk_random(size=(2, 16, 16), dtype=np.float64)
    expected = _naive_mask_apply(masks.astype(mask_dtype), data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=lambda: masks, mask_dtype=mask_dtype, mask_count=len(masks),
    )
    results = lt_ctx.run(analysis)

    assert results.mask_0.raw_data.dtype == mask_dtype

    assert np.allclose(
        results.mask_0.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1.raw_data,
        expected[1],
    )


def test_mask_udf(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = _mk_random(size=(16, 16))
    mask1 = sp.csr_matrix(_mk_random(size=(16, 16)))
    mask2 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
    # The ApplyMasksUDF returns data with shape ds.shape.nav + (mask_count, ),
    # different from ApplyMasksJob
    expected = np.moveaxis(_naive_mask_apply([mask0, mask1, mask2], data), (0, 1), (2, 0))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    udf = ApplyMasksUDF(
        mask_factories=[lambda: mask0, lambda: mask1, lambda: mask2]
    )
    results = lt_ctx.run_udf(udf=udf, dataset=dataset)

    assert np.allclose(results['intensity'].data, expected)


def test_numpy_is_sparse():
    mask = _mk_random(size=(16, 16))
    assert not is_sparse(mask)


def test_scipy_is_sparse():
    mask = sp.csr_matrix(_mk_random(size=(16, 16)))
    assert is_sparse(mask)


def test_sparse_is_sparse():
    mask = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
    assert is_sparse(mask)


def test_sparse_dok_is_sparse():
    mask = sparse.DOK.from_numpy(_mk_random(size=(16, 16)))
    assert is_sparse(mask)


def test_all_sparse_analysis(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = sp.csr_matrix(_mk_random(size=(16, 16)))
    mask1 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
    expected = _naive_mask_apply([mask0, mask1], data)

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1],
    )
    results = lt_ctx.run(analysis)

    assert np.allclose(
        results.mask_0.raw_data,
        expected[0],
    )
    assert np.allclose(
        results.mask_1.raw_data,
        expected[1],
    )


def _mask_from_analysis(dataset, analysis):
    slice_ = Slice(
        origin=(0, 0, 0),
        shape=Shape((1, 16, 16), sig_dims=2),
    )
    udf = analysis.get_udf()
    meta = UDFMeta(
        partition_slice=None,
        dataset_shape=dataset.shape,
        roi=None,
        dataset_dtype=dataset.dtype,
        input_dtype=dataset.dtype,
        corrections=None,
    )
    udf.set_meta(meta)
    return udf.masks.get(slice_, transpose=True)


def test_uses_sparse_all_default(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = sp.csr_matrix(_mk_random(size=(16, 16)))
    mask1 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1]
    )

    assert is_sparse(_mask_from_analysis(dataset, analysis))


def test_uses_sparse_mixed_default(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = sp.csr_matrix(_mk_random(size=(16, 16)))
    mask1 = _mk_random(size=(16, 16))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1]
    )

    assert not is_sparse(_mask_from_analysis(dataset, analysis))


def test_uses_sparse_true(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = _mk_random(size=(16, 16))
    mask1 = _mk_random(size=(16, 16))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1], use_sparse=True
    )

    assert is_sparse(_mask_from_analysis(dataset, analysis))


def test_uses_scipy_sparse(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = _mk_random(size=(16, 16))
    mask1 = _mk_random(size=(16, 16))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1], use_sparse='scipy.sparse'
    )

    assert sp.issparse(_mask_from_analysis(dataset, analysis))


def test_uses_sparse_pydata(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = _mk_random(size=(16, 16))
    mask1 = _mk_random(size=(16, 16))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1], use_sparse='sparse.pydata'
    )

    assert isinstance(_mask_from_analysis(dataset, analysis), sparse.SparseArray)


def test_uses_scipy_sparse_false(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = sp.csr_matrix(_mk_random(size=(16, 16)))
    mask1 = sp.csr_matrix(_mk_random(size=(16, 16)))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1], use_sparse=False
    )

    assert not is_sparse(_mask_from_analysis(dataset, analysis))


def test_uses_sparse_sparse_false(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
    mask0 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
    mask1 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))

    dataset = MemoryDataSet(data=data, tileshape=(4 * 4, 4, 4), num_partitions=2)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0, lambda: mask1], use_sparse=False
    )

    assert not is_sparse(_mask_from_analysis(dataset, analysis))


def test_masks_timeseries_2d_frames(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16), dtype="<u2")
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16, 16),
        num_partitions=2
    )
    mask0 = _mk_random(size=(16, 16))
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0],
    )
    results = lt_ctx.run(analysis)
    assert results.mask_0.raw_data.shape == (256,)


def test_masks_spectrum_linescan(lt_ctx):
    data = _mk_random(size=(16 * 16, 16 * 16), dtype="<u2")
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16 * 16),
        num_partitions=2,
        sig_dims=1,
    )
    mask0 = _mk_random(size=(16 * 16, ))
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0],
    )
    results = lt_ctx.run(analysis)
    assert results.mask_0.raw_data.shape == (16 * 16,)


def test_masks_spectrum(lt_ctx):
    data = _mk_random(size=(16, 16, 16 * 16), dtype="<u2")
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, 16 * 16),
        num_partitions=2,
        sig_dims=1,
    )
    mask0 = _mk_random(size=(16 * 16, ))
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0],
    )
    results = lt_ctx.run(analysis)
    assert results.mask_0.raw_data.shape == (16, 16)


@pytest.mark.slow
def test_masks_hyperspectral(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16, 16), dtype="<u2")
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16, 16),
        num_partitions=2,
        sig_dims=3,
    )
    mask0 = _mk_random(size=(16, 16, 16))
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0],
    )
    results = lt_ctx.run(analysis)
    assert results.mask_0.raw_data.shape == (16, 16)


def test_masks_complex_ds(lt_ctx, ds_complex):
    mask0 = _mk_random(size=(16, 16))
    analysis = lt_ctx.create_mask_analysis(
        dataset=ds_complex, factories=[lambda: mask0],
    )
    results = lt_ctx.run(analysis)
    assert results.mask_0.raw_data.shape == (16, 16)


def test_masks_complex_mask(lt_ctx, ds_complex):
    mask0 = _mk_random(size=(16, 16), dtype="complex64")
    analysis = lt_ctx.create_mask_analysis(
        dataset=ds_complex, factories=[lambda: mask0],
    )
    expected = _naive_mask_apply([mask0], ds_complex.data)
    results = lt_ctx.run(analysis)
    assert results.mask_0_complex.raw_data.shape == (16, 16)
    assert np.allclose(
        results.mask_0_complex.raw_data,
        expected
    )

    # also execute _run_mask_test_program to check sparse implementation.
    # _run_mask_test_program checks mask_0 result, which is np.abs(mask_0_complex)
    _run_mask_test_program(lt_ctx, ds_complex, mask0, np.abs(expected))


def test_numerics_fail(lt_ctx):
    dtype = 'float32'
    # Highest expected detector resolution
    RESOLUTION = 4096
    # Highest expected detector dynamic range
    RANGE = 1e6
    # default value for all cells
    VAL = 1.1

    data = np.full((2, 1, RESOLUTION, RESOLUTION), VAL, dtype=np.float32)
    data[0, 0, 0, 0] += VAL * RANGE
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, RESOLUTION, RESOLUTION),
        num_partitions=1,
        sig_dims=2,
    )
    mask0 = np.ones((RESOLUTION, RESOLUTION), dtype=np.float64)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0], mask_count=1, mask_dtype=dtype,
    )

    results = lt_ctx.run(analysis)
    expected = np.array([[
        [VAL*RESOLUTION**2 + VAL*RANGE],
        [VAL*RESOLUTION**2]
    ]])
    naive = _naive_mask_apply([mask0], data)
    naive_32 = _naive_mask_apply([mask0.astype(dtype)], data)
    # The masks are float64, that means the calculation is performed with high resolution
    # and the naive result should be correct
    assert np.allclose(expected, naive)
    # We make sure LiberTEM calculated this with the lower-precision dtype we set
    assert np.allclose(results.mask_0.raw_data, expected[0]) == np.allclose(naive_32, expected)
    # Confirm that the numerical precision is actually insufficient.
    # If this succeeds, we have to rethink the premise of this test.
    assert not np.allclose(results.mask_0.raw_data, expected[0])


def test_numerics_succeed(lt_ctx):
    dtype = 'float64'
    # Highest expected detector resolution
    RESOLUTION = 4096
    # Highest expected detector dynamic range
    RANGE = 1e6
    # default value for all cells
    VAL = 1.1

    data = np.full((2, 1, RESOLUTION, RESOLUTION), VAL, dtype=np.float32)
    data[0, 0, 0, 0] += VAL * RANGE
    dataset = MemoryDataSet(
        data=data,
        tileshape=(2, RESOLUTION, RESOLUTION),
        num_partitions=1,
        sig_dims=2,
    )
    mask0 = np.ones((RESOLUTION, RESOLUTION), dtype=np.float32)
    analysis = lt_ctx.create_mask_analysis(
        dataset=dataset, factories=[lambda: mask0], mask_count=1, mask_dtype=dtype,
    )

    results = lt_ctx.run(analysis)
    expected = np.array([[
        [VAL*RESOLUTION**2 + VAL*RANGE],
        [VAL*RESOLUTION**2]
    ]])
    naive = _naive_mask_apply([mask0.astype(dtype)], data.astype(dtype))

    assert np.allclose(expected, naive)
    assert np.allclose(expected[0], results.mask_0.raw_data)


@pytest.mark.parametrize(
    'kwargs', (
        {},
        {'use_sparse': True},
        {'use_sparse': False},
        {'use_sparse': 'sparse.pydata'},
        {'use_torch': True},
    )
)
@pytest.mark.parametrize(
    'backend', (None, NUMPY, CUPY)
)
def test_shifted_masks_constant_shifts(lt_ctx, kwargs, backend):
    with set_device_class(get_device_class(backend)):
        data = _mk_random(size=(16, 16, 16, 16), dtype="<u2")
        mask0 = _mk_random(size=(16, 16))
        mask1 = sp.csr_matrix(_mk_random(size=(16, 16)))
        mask2 = sparse.COO.from_numpy(_mk_random(size=(16, 16)))
        # The ApplyMasksUDF returns data with shape ds.shape.nav + (mask_count, ),
        expected = np.moveaxis(_naive_mask_apply(
            [mask0[0:15, 2:16], mask1[0:15, 2:16], mask2[0:15, 2:16]],
            data[..., 1:16, 0:14]
        ), (0, 1), (2, 0))

        dataset = MemoryDataSet(data=data, num_partitions=2)
        udf = ApplyMasksUDF(
            mask_factories=[lambda: mask0, lambda: mask1, lambda: mask2],
            shifts=np.round((1.1, -1.7)),
            **kwargs
        )
        if backend == CUPY and kwargs.get('use_sparse', False) in (True, 'sparse.pydata'):
            with pytest.raises(ValueError):
                # Implement like this so we can test the implementation
                # once CUPY + sparse.pydata is actually supported
                results = lt_ctx.run_udf(udf=udf, dataset=dataset)
            pytest.xfail('CuPy + sparse.pydata not yet supported')
        else:
            results = lt_ctx.run_udf(udf=udf, dataset=dataset)

        assert np.allclose(results['intensity'].data, expected)


def test_shifted_masks_scipy_sparse_raises(lt_ctx):
    with pytest.raises(ValueError):
        ApplyMasksUDF(
            mask_factories=lambda: np.ones((5, 6)),
            shifts=(1.2, -3.1),
            use_sparse='scipy.sparse',
        )


def _make_lambda(mask):
    return lambda: mask


def naive_shifted_mask_apply(
    masks: Sequence[np.ndarray],
    data: np.ndarray,
    shifts: np.ndarray,
) -> np.ndarray:
    """
    Manually compute the intersection of masks with data when
    shifts are applied, and return the dot product between them
    as an array of shape (num_frames, num_masks)

    Inputs must all be array-like with shapes
        masks: np.array(num_masks, h, w) or [num_masks * np.array(h, w)]
        data: (num_frames, h, w)
        shifts: (2,) or (num_frames, 2) of (y/x shifts of the mask)
    Sparse input masks or data are densified
    If shifts has shape (2,) then the same shift is applied to all frames/masks
    A zero-overlap between mask and frame returns zero
    for the dot product for that pair
    """
    # Densify inputs just in case!
    data = to_dense(data)
    masks = np.asarray([to_dense(m) for m in masks])
    num_frames, h, w = data.shape
    num_masks, mh, mw = masks.shape
    assert h == mh
    assert w == mw

    # Ensure we have one shift pair per frame even if constant shift
    if shifts.shape == (2,):
        shifts = np.repeat(shifts[np.newaxis, ...], num_frames, axis=0)
    assert shifts.shape == (num_frames, 2)

    expected = np.zeros((num_frames, num_masks), dtype=np.float64)
    for frame_idx, (dy, dx) in enumerate(shifts.astype(int)):
        # A positive shift value moves the mask down/right
        # so we need a slice from (0, 0) to (h-sy, w-sx)
        # A negative shift value moves the mask up/left
        # so we need a slice from (abs(sy), abs(sy)) to (h, w)
        # The min(max(...)) etc ensure we never slice beyond the
        # edge of the frame/mask in any situation
        my0 = min(max(0, -dy), h)  # the negative survives max(0, ...) only when dy is negative
        mx0 = min(max(0, -dx), w)  # the negative survives max(0, ...) only when dx is negative
        my1 = max(min(h, h - dy), 0)
        mx1 = max(min(w, w - dx), 0)
        # The frame is sliced in exactly the opposite way
        dy0 = min(max(0, dy), h)
        dx0 = min(max(0, dx), w)
        dy1 = max(min(h, h + dy), 0)
        dx1 = max(min(w, w + dx), 0)
        for mask_idx, mask in enumerate(masks):
            mask_sub = mask[np.s_[my0: my1, mx0: mx1]]
            if mask_sub.size == 0:
                # zero intersection, neutral element 0. already in expected
                continue
            data_sub = data[frame_idx][np.s_[dy0: dy1, dx0: dx1]]
            expected[frame_idx, mask_idx] = (mask_sub * data_sub).sum(axis=(-1, -2))
    return expected


def test_naive_shifted_apply():
    shifts = np.asarray([1, -2]).astype(int)
    mask_s = np.s_[0:15, 2:16]
    data_s = np.s_[:, 1:16, 0:14]
    data = _mk_random(size=(2, 16, 16))
    mask = _mk_random(size=(16, 16))
    assert np.allclose(
        naive_shifted_mask_apply([mask], data, shifts).squeeze(),
        (data[data_s] * mask[mask_s][np.newaxis, ...]).sum(axis=(-1, -2)),
    )


@pytest.mark.parametrize(
    'kwargs', (
        {},  # equivalent to use_sparse=None
        {'use_sparse': True},
        {'use_sparse': False},
        {'use_sparse': 'sparse.pydata'},
        {'use_torch': True},  # with use_sparse=None
    )
)
@pytest.mark.parametrize(
    'backend', (None, ) + ApplyMasksUDF(
        mask_factories=lambda: np.ones((1, 1, 1)), shifts=(1, -1),
    ).get_backends()
)
@pytest.mark.parametrize(
    'mask_types', [
        (np.asarray,),
        (sp.csr_matrix,),
        (np.asarray, sparse.COO.from_numpy),
        (sparse.COO.from_numpy, sp.csc_matrix,),
    ]
)
def test_shifted_masks_aux_shifts(lt_ctx, kwargs, backend, mask_types):
    with set_device_class(get_device_class(backend)):
        # nonsquare frames to test ordering is correct
        num_frames, h, w = 2, 18, 12
        data = np.random.uniform(size=(num_frames, h, w)).astype(np.float32)
        if backend in SPARSE_BACKENDS:
            data = _mk_random((num_frames, h, w), array_backend=SPARSE_COO, sparse_density=0.1)
        else:
            data = _mk_random((num_frames, h, w), array_backend=NUMPY)
        shifts = np.random.uniform(
            low=-5,
            high=5,
            size=(num_frames, 2),
        )

        masks = []
        mask_factories = []
        for mask_type in mask_types:
            mask = mask_type(_mk_random(size=(h, w)))
            masks.append(mask)
            mask_factories.append(_make_lambda(mask))

        expected = naive_shifted_mask_apply(masks, data, shifts)
        dataset = MemoryDataSet(
            data=data,
            array_backends=(backend, ) if backend is not None else None
        )
        udf = ApplyMasksUDF(
            mask_factories=mask_factories,
            shifts=ApplyMasksUDF.aux_data(
                data=shifts.ravel(),
                kind='nav',
                extra_shape=(2, ),
                dtype=float,
            ),
            **kwargs,
        )

        use_sparse = kwargs.get('use_sparse', None)  # default value
        # if forcing sparse on CuPy, raise, else densify to process
        if backend in CUPY_BACKENDS and use_sparse in (True, 'sparse.pydata'):
            with pytest.raises(ValueError):
                # Implement like this so we can test the implementation
                # once CUPY + sparse.pydata is actually supported
                results = lt_ctx.run_udf(udf=udf, dataset=dataset)
            pytest.xfail('CuPy + sparse.pydata not yet supported')
        else:
            results = lt_ctx.run_udf(udf=udf, dataset=dataset)

        assert np.allclose(results['intensity'].data, expected)


def test_shifted_masks_zero_overlap(lt_ctx):
    num_frames, h, w = 2, 18, 12
    data = _mk_random(size=(num_frames, h, w))

    dataset = MemoryDataSet(data=data)
    udf = ApplyMasksUDF(
        mask_factories=[lambda: _mk_random(size=(h, w))],
        shifts=(-20, 15),
    )

    results = lt_ctx.run_udf(udf=udf, dataset=dataset)
    assert np.allclose(results['intensity'].data, 0.)


@pytest.mark.parametrize(
    'backend', (None, NUMPY, CUPY)
)
def test_shifted_masks_stacked(lt_ctx, backend):
    with set_device_class(get_device_class(backend)):
        # nonsquare frames to test ordering is correct
        num_frames, h, w = 2, 18, 12
        data = _mk_random(size=(num_frames, h, w))
        shifts = np.random.uniform(
            low=-5.,
            high=5.,
            size=(num_frames, 2),
        )

        masks = _mk_random(size=(3, h, w))

        expected = naive_shifted_mask_apply(masks, data, shifts)
        dataset = MemoryDataSet(data=data)
        udf = ApplyMasksUDF(
            mask_factories=lambda: masks,
            shifts=ApplyMasksUDF.aux_data(
                data=shifts.ravel(),
                kind='nav',
                extra_shape=(2, ),
                dtype=float,
            ),
        )

        results = lt_ctx.run_udf(udf=udf, dataset=dataset)
        assert np.allclose(results['intensity'].data, expected)


def test_shifted_masks_descan(lt_ctx):
    # simulate a descan error on a constant frame
    # check that unshifted ApplyMasksUDF has aliasing effects and that
    # applying the shift values gives the correct constant result
    h, w = (9, 9)
    frame = np.zeros((h, w), dtype=int)
    frame = circular(4, 4, w, h, 1)
    frame_sum = frame.sum()
    frames = []
    shifts = np.moveaxis(np.mgrid[-2:4:2, -2:4:2], 0, -1)
    for yroll, xroll in shifts.reshape(-1, 2):
        frames.append(np.roll(frame, (yroll, xroll), axis=(0, 1)))
    data = np.stack(frames, axis=0)
    ds = lt_ctx.load('memory', data=data)
    mask = circular(4, 4, w, h, 2)
    udf = ApplyMasksUDF(mask_factories=[lambda: mask])
    results = lt_ctx.run_udf(udf=udf, dataset=ds)
    assert not (results['intensity'].data == frame_sum).all()
    assert results['intensity'].data.reshape(3, 3)[1, 1] == frame_sum
    udf = ApplyMasksUDF(
        mask_factories=[lambda: mask],
        shifts=ApplyMasksUDF.aux_data(
            data=shifts.ravel(),
            kind='nav',
            extra_shape=(2, ),
            dtype=int,
        ),
    )
    results = lt_ctx.run_udf(udf=udf, dataset=ds)
    assert (results['intensity'].data == frame_sum).all()


def test_backend_unknown():
    with pytest.raises(ValueError):
        ApplyMasksUDF(
            mask_factories=[lambda: np.ones((5, 5))],
            backends=('BACKEND_DOES_NOT_EXIST',)
        )



from typing import Union
from collections.abc import Iterable
import itertools
import warnings


def assign_cudas(cudas: Union[int, Iterable[int]]) -> Iterable[int]:
    """
    Takes the cudas argument to :code:`cluster_spec` and
    converts it into a properly formatted iterable of CUDA
    device ids

    If cudas is an integer, assigns this many workers to
    device ids in a round-robin fashion, where CUDA devices
    can be detected. If the devices cannot be detected raise
    a warning and assign device_ids in a sequential fashion.

    Will also raise a warning if cudas is a non-empty iterable
    on a system where the CUDA devices cannot be detected.
    """
    if isinstance(cudas, int) or len(cudas):
        # Needed to know if we can assign CUDA workers
        from libertem.utils.devices import detect
        avail_cudas = detect()['cudas']
        if not avail_cudas and cudas:  # needed in case cudas == 0
            warnings.warn('Specifying CUDA workers on system with '
                          'no visible CUDA devices',
                          RuntimeWarning)
            # If we are assigning from int, just use increasing
            # device indices even if they are unavailable
            avail_cudas = itertools.count()

        if isinstance(cudas, int):
            # Round-Robin-assign to available CUDA devices
            # Can override by specifying cudas as an iterable
            cudas_iter = itertools.cycle(avail_cudas)
            cudas = tuple(next(cudas_iter) for _ in range(cudas))

    return cudas



import numpy as np
import pytest

from libertem.analysis.raw import PickFrameAnalysis
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


def test_pick_analysis(lt_ctx):
    """
    the other tests cover the pick job, this one uses the analysis
    """
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 5, "y": 5})
    result = lt_ctx.run(analysis)

    assert result.intensity.raw_data.shape == (16, 16)
    assert np.allclose(result.intensity.raw_data, data[5, 5])
    assert np.allclose(result.intensity_lin.raw_data, data[5, 5])


def test_pick_from_3d_ds(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 5})
    result = lt_ctx.run(analysis)

    assert result.intensity.raw_data.shape == (16, 16)
    assert np.allclose(result.intensity.raw_data, data[5])
    assert np.allclose(result.intensity_lin.raw_data, data[5])


def test_pick_analysis_via_api_1(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = lt_ctx.create_pick_analysis(dataset=dataset, x=8, y=7)
    result = lt_ctx.run(analysis)

    assert result.intensity.raw_data.shape == (16, 16)
    assert np.allclose(result.intensity.raw_data, data[7, 8])
    assert np.allclose(result.intensity_lin.raw_data, data[7, 8])


def test_pick_analysis_via_api_2_3d_ds(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = lt_ctx.create_pick_analysis(dataset=dataset, x=8)
    result = lt_ctx.run(analysis)

    assert result.intensity.raw_data.shape == (16, 16)
    assert np.allclose(result.intensity.raw_data, data[8])
    assert np.allclose(result.intensity_lin.raw_data, data[8])


def test_pick_analysis_via_api_3_3d_ds_fail_1(lt_ctx):
    data = _mk_random(size=(16 * 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8, "z": 11})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)


def test_pick_analysis_via_api_3_3d_ds_fail_2(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8, "z": 11})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)


def test_pick_analysis_via_api_3_3d_ds_fail_3(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)


def test_pick_analysis_via_api_3_3d_ds_fail_4(lt_ctx):
    data = _mk_random(size=(16, 16, 16, 16, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8, "z": 11})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)


def test_pick_analysis_via_api_3_3d_ds_fail_5(lt_ctx):
    data = _mk_random(size=(16, 256, 16, 16))
    dataset = MemoryDataSet(
        data=data,
        tileshape=(1, 16, 16),
        num_partitions=2,
        sig_dims=2
    )

    analysis = PickFrameAnalysis(dataset=dataset, parameters={"x": 7, "y": 8, "z": 11})
    with pytest.raises(ValueError):
        lt_ctx.run(analysis)



import os
import stat

import tornado.web

from libertem.io.fs import get_fs_listing, FSError, stat_path
from .messages import Message
from .state import SharedState


class LocalFSBrowseHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry):
        self.state = state
        self.event_registry = event_registry

    async def get(self):
        executor = await self.state.executor_state.get_executor()
        path = self.request.arguments['path']
        assert len(path) == 1
        path = path[0].decode("utf8")
        try:
            with self.state.executor_state.keep_alive():
                listing = await executor.run_function(get_fs_listing, path)
            msg = Message().directory_listing(
                **listing
            )
            self.write(msg)
        except FSError as e:
            msg = Message().browse_failed(
                path=path,
                code=e.code,
                msg=str(e),
                alternative=e.alternative,
            )
            self.write(msg)


class LocalFSStatHandler(tornado.web.RequestHandler):
    def initialize(self, state: SharedState, event_registry) -> None:
        self.state = state
        self.event_registry = event_registry

    async def get(self):
        executor = await self.state.executor_state.get_executor()
        path = self.request.arguments['path']
        assert len(path) == 1
        path = path[0].decode("utf8")
        try:
            stat_result = await executor.run_function(stat_path, path)
            if stat.S_ISDIR(stat_result.st_mode) and not path.endswith(os.path.sep):
                path_slash = path + os.path.sep
            else:
                path_slash = path
            msg = Message().browse_stat_result(
                path=path,
                dirname=os.path.dirname(path_slash),
                basename=os.path.basename(path_slash),
                stat_result=stat_result,
            )
            self.write(msg)
        except FSError as e:
            msg = Message().stat_failed(
                path=path,
                code=e.code,
                msg=str(e),
                alternative=e.alternative,
            )
            self.write(msg)



import numpy as np
import pytest

from libertem.api import Context
from libertem.common.shape import Shape
from libertem.udf.base import UDF, MergeAttrMapping
from libertem.io.dataset.memory import MemoryDataSet
from libertem.common.buffers import (
    BufferWrapper, InvalidMaskError, get_inner_slice, get_bbox, get_bbox_slice,
)
from libertem.common.math import count_nonzero, prod


class ValidNavMaskUDF(UDF):
    def __init__(self, debug=True):
        super().__init__(debug=debug)

    def get_result_buffers(self):
        return {
            'buf_sig': self.buffer(kind='sig', dtype=np.float32),
            'buf_nav': self.buffer(kind='nav', dtype=np.float32),
            'buf_single': self.buffer(kind='single', dtype=np.float32, extra_shape=(1,)),
        }

    def get_results(self):
        assert self.meta.get_valid_nav_mask() is not None
        assert self.meta.get_valid_nav_mask().sum() > 0, \
            "get_results is not called with an empty valid nav mask"
        assert len(self.meta.get_valid_nav_mask().shape) == 1, \
            "valid_nav_mask should be flattened"
        if self.meta.roi is not None:
            assert self.meta.get_valid_nav_mask().shape[0] == np.count_nonzero(self.meta.roi), \
                "if a `roi` is given, the valid nav mask should be compressed to it by default"
            full_mask = self.meta.get_valid_nav_mask(full_nav=True)
            assert full_mask.shape[0] == self.meta.dataset_shape.nav.size, \
                "when passing `full_nav=True`, the shape must match the flattened ds shape"
        if self.params.debug:
            print("get_results", self.meta.get_valid_nav_mask())
        results = super().get_results()
        return results

    def process_frame(self, frame):
        assert self.meta.get_valid_nav_mask() is None
        assert self.meta.get_valid_nav_mask(full_nav=True) is None
        self.results.buf_sig += frame
        self.results.buf_nav[:] = frame.sum()
        self.results.buf_single[:] = frame.sum()

    def merge(self, dest, src):
        assert self.meta.get_valid_nav_mask() is not None
        assert not np.allclose(True, self.meta.get_valid_nav_mask()), \
            "valid nav mask should be the already merged positions! can't be all-True"
        if self.params.debug:
            print("merge", self.meta.get_valid_nav_mask())
        dest.buf_sig += src.buf_sig
        dest.buf_single += src.buf_single
        dest.buf_nav[:] = src.buf_nav


def test_valid_nav_mask_available(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=ValidNavMaskUDF()):
        # TODO: maybe compare damage we got in `get_results` with `res.damage` here?
        pass


@pytest.mark.parametrize("with_roi", [True, False])
def test_valid_nav_mask_delayed(delayed_executor, with_roi: bool):
    ctx = Context(executor=delayed_executor)
    if with_roi:
        roi = np.zeros((16, 16), dtype=bool)
        roi[4:-4, 4:-4] = True
    else:
        roi = None
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    for res in ctx.run_udf_iter(dataset=dataset, udf=ValidNavMaskUDF(), roi=roi):
        res.buffers[0]['buf_nav'].data
        assert np.allclose(
            res.buffers[0]['buf_nav'].valid_mask,
            res.damage,
        )


def test_valid_nav_mask_available_roi(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    roi = np.zeros((16, 16), dtype=bool)
    roi[4:-4, 4:-4] = True
    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=ValidNavMaskUDF(debug=False), roi=roi):
        print("damage", res.damage.data)
        print("raw damage", res.damage.raw_data)


def test_valid_nav_mask_available_random_roi(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    roi = np.random.choice([True, False], size=(16, 16))
    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=ValidNavMaskUDF(debug=False), roi=roi):
        print("damage", res.damage.data)
        print("raw damage", res.damage.raw_data)


class AdjustValidMaskUDF(UDF):
    def get_result_buffers(self):
        return {
            'all_valid': self.buffer(kind='sig', dtype=np.float32),
            'all_invalid': self.buffer(kind='sig', dtype=np.float32),
            'keep': self.buffer(kind='nav', dtype=np.float32),
            'nav_with_extra': self.buffer(kind='nav', dtype=np.float32, extra_shape=(2,)),
            'custom_2d': self.buffer(kind='single', dtype=np.float32, extra_shape=(64, 64)),
        }

    def get_results(self):
        custom_mask = np.zeros((64, 64), dtype=bool)
        custom_mask[:, 32:] = True

        return {
            'all_valid': self.with_mask(self.results.all_valid, mask=True),
            'all_invalid': self.with_mask(self.results.all_invalid, mask=False),
            'keep': self.results.keep,
            'nav_with_extra': self.results.nav_with_extra,
            'custom_2d': self.with_mask(self.results.custom_2d, mask=custom_mask),
        }

    def process_frame(self, frame):
        self.results.all_valid += frame
        self.results.all_invalid += frame
        self.results.keep[:] = frame.sum()
        self.results.nav_with_extra[:] = frame.sum()
        self.results.custom_2d[:] = 42

    def merge(self, dest, src):
        dest.all_valid += src.all_valid
        dest.all_invalid += src.all_invalid
        dest.custom_2d += src.custom_2d
        dest.keep[:] = src.keep
        dest.nav_with_extra[:] = src.nav_with_extra


@pytest.mark.parametrize(
    "with_roi", [True, False]
)
@pytest.mark.parametrize(
    "executor", ["inline", "delayed"],
)
def test_adjust_valid_mask(with_roi: bool, executor: str, delayed_executor, lt_ctx):
    """
    Test that we can adjust the valid mask in `get_results`
    """
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    if executor == "inline":
        ctx = lt_ctx
    else:
        ctx = Context(executor=delayed_executor)

    if with_roi:
        roi = np.random.choice([True, False], size=dataset.shape.nav)
    else:
        roi = None

    custom_expected = np.zeros((64, 64), dtype=bool)
    custom_expected[:, 32:] = True

    for res in ctx.run_udf_iter(dataset=dataset, udf=AdjustValidMaskUDF(), roi=roi):
        # invariants that hold for any intermediate results:
        # all-valid result:
        valid_mask = res.buffers[0]['all_valid'].valid_mask
        assert np.allclose(valid_mask, True)
        assert valid_mask.shape == res.buffers[0]['all_valid'].data.shape

        # all-invalid result:
        invalid_mask = res.buffers[0]['all_invalid'].valid_mask
        assert np.allclose(invalid_mask, False)
        assert invalid_mask.shape == res.buffers[0]['all_invalid'].data.shape

        # same as "damage", default for kind='nav' buffers:
        keep_mask = res.buffers[0]['keep'].valid_mask
        assert np.allclose(keep_mask, res.damage.data)

        # nav with extra_shape:
        extra_mask = res.buffers[0]['nav_with_extra'].valid_mask
        assert np.allclose(
            extra_mask,
            res.damage.data.reshape((16, 16, 1)),  # broadcastable to nav+extra
        )

        # custom 2d mask:
        custom_mask = res.buffers[0]['custom_2d'].valid_mask
        assert np.allclose(custom_mask, custom_expected)


class CustomMaskFromParams(UDF):
    def __init__(self, mask):
        super().__init__(mask=mask)

    def get_result_buffers(self) -> dict[str, BufferWrapper]:
        return {
            'custom': self.buffer(kind='single', dtype='float32', extra_shape=(64, 64, 3)),
        }

    def get_results(self):
        return {
            'custom': self.with_mask(np.zeros((64, 64, 3), dtype="float32"), mask=self.params.mask),
        }

    def process_frame(self, frame):
        pass

    def merge(self, dest, src):
        pass


@pytest.mark.parametrize("mask_shape", [
    (32, 32),
    (1, 32),
    (64, 64),  # needs to be (64, 64, 1) to be able to broadcast
    (64, 64, 4),
    (1, 1, 4),
    (1, 1, 1, 1),  # that's too many...
])
def test_custom_mask_invalid_shape(mask_shape, lt_ctx):
    """
    Examples of mask shapes that are incompatible with the 'custom' buffer
    defined above. Make sure these raise an appropriate exception.
    """
    dataset = MemoryDataSet(datashape=[16, 16, 4, 4], num_partitions=4)

    mask = np.zeros(mask_shape, dtype=bool)

    with pytest.raises(InvalidMaskError):
        for res in lt_ctx.run_udf_iter(dataset=dataset, udf=CustomMaskFromParams(mask=mask)):
            pass


@pytest.mark.parametrize("mask_dtype", [
    int,
    "float32",
    "complex64",
    # ...
])
def test_custom_mask_invalid_dtype(mask_dtype, lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 4, 4], num_partitions=4)
    mask = np.zeros((16, 16), dtype=mask_dtype)

    with pytest.raises(InvalidMaskError):
        for res in lt_ctx.run_udf_iter(dataset=dataset, udf=CustomMaskFromParams(mask=mask)):
            pass


@pytest.mark.parametrize("mask_shape", [
    (),
    (1,),
    (1, 1),
    (1, 1, 1),
    (64, 64, 1),
    (64, 64, 3),
])
def test_custom_mask_valid(mask_shape, lt_ctx):
    """
    Examples of mask shapes that should be compatible with the shape of the 'custom'
    buffer defined in the `CustomMaskFromParams` UDF.
    """
    dataset = MemoryDataSet(datashape=[16, 16, 4, 4], num_partitions=4)

    mask = np.zeros(mask_shape, dtype=bool)

    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=CustomMaskFromParams(mask=mask)):
        pass


def test_valid_mask_slice_bounding(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)

    custom_expected = np.zeros((64, 64), dtype=bool)
    custom_expected[:, 32:] = True

    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=AdjustValidMaskUDF()):
        # invariants that hold for any intermediate results:
        # all-valid result:
        buf = res.buffers[0]['all_valid']
        assert buf.data[buf.valid_slice_bounding].shape == buf.data.shape

        # all-invalid result:
        buf = res.buffers[0]['all_invalid']
        assert prod(buf.data[buf.valid_slice_bounding].shape) == 0

        # same as "damage", default for kind='nav' buffers:
        buf = res.buffers[0]['keep']
        assert prod(buf.data[buf.valid_slice_bounding].shape) >= np.count_nonzero(res.damage.data)

        # custom 2d mask:
        buf = res.buffers[0]['custom_2d']
        assert buf.valid_slice_bounding == np.s_[0:64, 32:64]


def test_masked_data(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)

    custom_expected = np.zeros((64, 64), dtype=bool)
    custom_expected[:, 32:] = True

    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=AdjustValidMaskUDF()):
        for k, buf in res.buffers[0].items():
            # "checksum" over accessing via data[valid_mask] vs. masked_data
            masked_sum = buf.masked_data.sum()  # a value or the marker `masked`

            # if everything is masked out, the masked sum is just the `masked`
            # marker, which is not equal to zero:
            assert np.sum(buf.data[buf.valid_mask]) == masked_sum\
                or np.allclose(buf.masked_data.mask, True)


def test_raw_masked_data(lt_ctx):
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)

    roi = np.random.choice(a=[True, False], size=(16, 16))

    custom_expected = np.zeros((64, 64), dtype=bool)
    custom_expected[:, 32:] = True

    for res in lt_ctx.run_udf_iter(dataset=dataset, udf=AdjustValidMaskUDF(), roi=roi):
        for k, buf in res.buffers[0].items():
            # "checksum" over accessing via data[valid_mask] vs. masked_data
            masked_sum = buf.raw_masked_data.sum()  # a value or the marker `masked`

            # if everything is masked out, the masked sum is just the `masked`
            # marker, which is not equal to zero:
            assert np.sum(buf.data[buf.valid_mask]) == masked_sum \
                or np.allclose(buf.raw_masked_data.mask, True)

            # 'nav' is the interesting case where the `roi` is affecting things:
            if buf.kind == 'nav':
                matched_roi = roi.reshape(roi.shape + (1,) * len(buf.extra_shape))
                matched_roi = np.broadcast_to(matched_roi, buf.valid_mask.shape)
                matched_roi = matched_roi.reshape(roi.shape + buf.extra_shape)
                assert np.count_nonzero(~buf.raw_masked_data.mask) == \
                    np.count_nonzero(matched_roi[buf.valid_mask])

            assert np.sum(buf.raw_data[buf._valid_mask]) == masked_sum\
                or np.allclose(buf.raw_masked_data.mask, True)


def test_get_inner_slice():
    a = np.zeros((16, 16), dtype=bool)
    a[5:7] = 1
    a[8, 8] = 1
    a[-1, -1] = 1
    assert get_inner_slice(a, axis=0) == np.s_[5:7, :]

    b = np.zeros((16, 16, 16), dtype=bool)
    b[5:7] = 1
    b[8, 1] = 1
    b[-1, -1] = 1
    assert get_inner_slice(b, axis=0) == np.s_[5:7, :, :]

    c = np.zeros((16, 16, 16), dtype=bool)
    c[:, 5:7, :] = 1
    assert get_inner_slice(c, axis=1) == np.s_[:, 5:7, :]


@pytest.mark.with_numba
def test_get_bbox():
    a = np.zeros((16, 16), dtype=bool)
    a[6, 6] = 1
    assert get_bbox(a) == (6, 6, 6, 6)

    a = np.zeros((16, 16, 16), dtype=bool)
    a[:, 6, 6] = 1
    assert get_bbox(a) == (0, 15, 6, 6, 6, 6)


def test_get_bbox_slice():
    a = np.zeros((16, 16), dtype=bool)
    a[6, 6] = 1
    assert get_bbox_slice(a) == np.s_[6:7, 6:7]

    a = np.zeros((16, 16, 16), dtype=bool)
    a[:, 6, 6] = 1
    assert get_bbox_slice(a) == np.s_[0:16, 6:7, 6:7]

    a = np.zeros((16, 16, 16), dtype=bool)
    a[:, 6, 6] = 1
    a[:, -1, -1] = 1
    assert get_bbox_slice(a) == np.s_[0:16, 6:16, 6:16]


def test_default_mask_nav_extra_shape():
    buf = BufferWrapper(kind='nav', extra_shape=(1, 2), dtype="float32")
    valid_nav_mask = np.array([1, 1, 0], dtype=bool)
    ds_shape = Shape((3, 1, 16, 16), sig_dims=2)
    nav_mask = buf.make_default_mask(
        valid_nav_mask=valid_nav_mask, dataset_shape=ds_shape, roi=None,
    )
    assert nav_mask.shape == (3, 1, 2)
    assert np.allclose(nav_mask, valid_nav_mask.reshape((3, 1, 1)))


def test_default_mask_nav_extra_shape_with_roi():
    buf = BufferWrapper(kind='nav', extra_shape=(1, 2), dtype="float32")
    valid_nav_mask = np.array([1, 0], dtype=bool)
    roi = np.array([True, False, True])
    ds_shape = Shape((3, 1, 16, 16), sig_dims=2)
    nav_mask = buf.make_default_mask(
        valid_nav_mask=valid_nav_mask, dataset_shape=ds_shape, roi=roi,
    )
    assert nav_mask.shape == (2, 1, 2)
    assert np.allclose(nav_mask, valid_nav_mask.reshape((2, 1, 1)))


def test_default_mask_sig_extra_shape():
    buf = BufferWrapper(kind='sig', extra_shape=(1, 2), dtype="float32")
    valid_nav_mask = np.array([1, 1, 0], dtype=bool)
    ds_shape = Shape((3, 1, 32, 32), sig_dims=2)
    sig_mask = buf.make_default_mask(
        valid_nav_mask=valid_nav_mask, dataset_shape=ds_shape, roi=None,
    )
    assert sig_mask.shape == (32, 32, 1, 2)
    assert np.allclose(sig_mask, 1)


def test_default_mask_single_extra_shape():
    buf = BufferWrapper(kind='single', extra_shape=(1, 2), dtype="float32")
    valid_nav_mask = np.array([1, 1, 0], dtype=bool)
    ds_shape = Shape((3, 1, 32, 32), sig_dims=2)
    sig_mask = buf.make_default_mask(
        valid_nav_mask=valid_nav_mask, dataset_shape=ds_shape, roi=None,
    )
    assert sig_mask.shape == (1, 2)
    assert np.allclose(sig_mask, 1)


class CustomValidMask(UDF):
    def get_result_buffers(self):
        nav_shape = self.meta.dataset_shape.nav

        if self.meta.roi is not None:
            nav_shape = (count_nonzero(self.meta.roi),)

        return {
            'custom_2d': self.buffer(kind='single', dtype=np.float32, extra_shape=nav_shape),
        }

    def process_frame(self, frame):
        self.results.custom_2d[self.meta.coordinates] = np.sum(frame)

    def get_results(self):
        # in this case, we set the valid mask of the kind=single buffer to
        # match what a nav buffer would do.
        valid_nav_mask = self.meta.get_valid_nav_mask()

        return {
            'custom_2d': self.with_mask(
                self.results.custom_2d,
                mask=valid_nav_mask.reshape(self.results.custom_2d.shape)
            ),
        }

    def merge(self, dest: MergeAttrMapping, src: MergeAttrMapping):
        dest.custom_2d += src.custom_2d


@pytest.mark.parametrize(
    "with_roi", [True, False]
)
@pytest.mark.parametrize(
    "executor", ["inline", "delayed"],
)
def test_adjust_valid_mask_extra(with_roi: bool, executor: str, delayed_executor, lt_ctx):
    """
    Test that we can adjust the valid mask in `get_results`
    """
    dataset = MemoryDataSet(datashape=[16, 16, 32, 32], num_partitions=4)
    if executor == "inline":
        ctx = lt_ctx
    else:
        ctx = Context(executor=delayed_executor)

    if with_roi:
        roi = np.random.choice([True, False], size=dataset.shape.nav)
    else:
        roi = None

    ctx.run_udf(dataset=dataset, udf=CustomValidMask(), roi=roi)



import os
import logging

from ncempy.io.mrc import fileMRC

from libertem.common.math import prod, make_2D_square
from libertem.common import Shape
from libertem.common.messageconverter import MessageConverter
from .base import DataSet, FileSet, BasePartition, DataSetException, DataSetMeta, File
from .base.backend import IOBackend
from .base.backend_mmap import MMapBackendImpl, MMapFileBase

log = logging.getLogger(__name__)


class MRCDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/MRCDatasetParams.schema.json",
        "title": "MRCDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "MRC"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
        },
        "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class MRCBackendFile(MMapFileBase):
    def __init__(self, path, desc):
        self.path = path
        self.desc = desc
        self._handle = None
        self._mmap = None

    def open(self):
        self._handle = fileMRC(self.path)
        self._mmap = self._handle.getMemmap()
        return self

    def close(self):
        self._handle = None
        self._mmap = None

    @property
    def array(self):
        return self._mmap

    @property
    def mmap(self):
        return self._mmap


class MRCBackend(IOBackend):
    def get_impl(self):
        return MRCBackendImpl()


class MRCBackendImpl(MMapBackendImpl):
    FILE_CLS = MRCBackendFile


class MRCDataSet(DataSet):
    """
    Read MRC files.

    Examples
    --------

    >>> ds = ctx.load("mrc", path="/path/to/file.mrc")  # doctest: +SKIP

    Parameters
    ----------
    path: str
        Path to the .mrc file

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
    """
    def __init__(self, path, nav_shape=None, sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        if io_backend is not None:
            raise ValueError("MRCDataSet currently doesn't support alternative I/O backends")
        self._path = path
        self._meta = None
        self._filesize = None
        self._image_count = None
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset

    def _do_initialize(self):
        self._filesize = os.stat(self._path).st_size
        f = fileMRC(self._path)
        data = f.getMemmap()
        native_shape = data.shape
        dtype = data.dtype

        self._image_count = native_shape[0]

        if self._nav_shape is None:
            self._nav_shape = tuple((int(native_shape[0]),))

        native_sig_shape = tuple(
            int(i)
            for i in f.gridSize
            if i != 1
        )
        if self._sig_shape is None:
            self._sig_shape = native_sig_shape
        elif int(prod(self._sig_shape)) != int(prod(native_sig_shape)):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(native_sig_shape))
            )

        self._sig_dims = len(self._sig_shape)
        self._shape = Shape(self._nav_shape + self._sig_shape, sig_dims=self._sig_dims)
        self._nav_shape_product = self._shape.nav.size
        self._sync_offset_info = self.get_sync_offset_info()

        self._meta = DataSetMeta(
            shape=self._shape,
            raw_dtype=dtype,
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    def get_diagnostics(self):
        return [
            {"name": "dtype", "value": str(self._meta.raw_dtype)},
        ]

    @classmethod
    def get_msg_converter(cls):
        return MRCDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"mrc"}

    @classmethod
    def detect_params(cls, path, executor):
        if path.lower().endswith(".mrc"):
            f = fileMRC(path)
            data = f.getMemmap()
            shape = data.shape
            sig_shape = tuple(
                int(i)
                for i in f.gridSize
                if i != 1
            )
            nav_shape = shape[0]
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": make_2D_square((int(nav_shape),)),
                    "sig_shape": sig_shape,
                },
                "info": {
                    "image_count": int(nav_shape),
                    "native_sig_shape": sig_shape,
                }
            }
        return False

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    def check_valid(self):
        return True  # anything to check?

    def get_cache_key(self):
        return {
            "path": self._path,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def _get_fileset(self):
        assert self._image_count is not None
        return FileSet([
            File(
                path=self._path,
                start_idx=0,
                end_idx=self._image_count,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
            )
        ])

    @classmethod
    def get_supported_io_backends(self):
        return []

    def get_io_backend(self):
        return MRCBackend()

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in MRCPartition.make_slices(
                shape=self.shape,
                num_partitions=self.get_num_partitions(),
                sync_offset=self._sync_offset):
            yield MRCPartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<MRCDataSet for {self._path}>"


class MRCPartition(BasePartition):
    pass



import numpy as np
import scipy.sparse as sp
import sparse
from typing import Union, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from libertem.common.shape import Shape


def to_dense(a):
    if isinstance(a, sparse.SparseArray):
        return a.todense()
    elif sp.issparse(a):
        return a.toarray()
    else:
        return np.array(a)


def to_sparse(a, shape: Optional[Union['Shape', tuple[int, ...]]] = None):
    if isinstance(a, sparse.COO):
        return a
    elif isinstance(a, sparse.SparseArray):
        return sparse.COO(a)
    elif sp.issparse(a):
        return sparse.COO.from_scipy_sparse(a)
    elif isinstance(a, (tuple, list)):
        if all(isinstance(aa, int) for aa in a):
            a = ((a, True),)
        unique = {aa[-1] for aa in a}
        if len(unique) != 1:
            raise ValueError('Cannot cast iterable roi coords with '
                             f'more than one truth value {unique}')
        roi_val = bool(tuple(unique)[0])
        fill_val = not roi_val
        return sparse.COO.from_iter(a, shape=tuple(shape), fill_value=fill_val, dtype=bool)
    else:
        return sparse.COO.from_numpy(np.array(a))


def sparse_to_coo(a, shape: Optional[Union['Shape', tuple[int, ...]]] = None):
    if a is None or isinstance(a, np.ndarray):
        return a
    return to_sparse(a, shape=shape)


def is_sparse(a):
    return isinstance(a, sparse.SparseArray) or sp.issparse(a)


def assert_sparse(a) -> sparse.SparseArray:
    assert is_sparse(a)
    return a



import operator
import functools
from typing import Any, Union, overload
from collections.abc import Iterator, Sequence


class Shape:
    """
    Create a Shape that knows how many dimensions are part of navigation/signal.
    It is assumed that the signal is in the last `sig_dims` dimensions.

    Parameters
    ----------
    shape : tuple of int
        the shape we want to work with, as n-tuple (like numpy array shapes)
    sig_dims : int
        the number of dimensions that are considered part of the signal
    """

    __slots__ = ["_sig_dims", "_nav_shape", "_sig_shape"]

    def __init__(self, shape: "ShapeLike", sig_dims: int):
        nav_dims = len(shape) - sig_dims
        self._sig_dims = sig_dims
        self._nav_shape = tuple(shape)[:nav_dims]
        self._sig_shape = tuple(shape)[nav_dims:]

    @property
    def nav(self) -> "Shape":
        """
        Crop to navigation dimensions

        #TODO Should be refactored with functools.cached_property when supported

        Returns
        -------
        shape : Shape
            like this shape, but without the signal dimensions

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((5, 5, 16, 16), sig_dims=2)
        >>> s.nav
        (5, 5)
        """
        return NavOnlyShape(shape=self._nav_shape)

    @property
    def sig(self) -> "Shape":
        """
        Crop to signal dimensions

        #TODO Should be refactored with functools.cached_property when supported

        Returns
        -------
        shape : Shape
            like this shape, but without the navigation dimensions

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((5, 5, 16, 16), sig_dims=2)
        >>> s.sig
        (16, 16)
        """
        return SigOnlyShape(shape=self._sig_shape)

    def to_tuple(self) -> tuple[int, ...]:
        return tuple(self)

    @property
    def size(self) -> int:
        """
        Number of elements covered by this shape

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((16, 16), sig_dims=2)
        >>> s.size
        256
        """
        shape_tuple = self.to_tuple()
        if len(shape_tuple) == 0:
            return 0
        return int(functools.reduce(operator.mul, shape_tuple))

    def flatten_nav(self) -> "Shape":
        """
        Returns a new Shape that is flat in the navigation dimensions

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((5, 5, 16, 16), sig_dims=2)
        >>> s.flatten_nav()
        (25, 16, 16)
        """
        return Shape(shape=(self.nav.size,) + self._sig_shape, sig_dims=self._sig_dims)

    def flatten_sig(self) -> "Shape":
        """
        Flatten in the signal dimensions

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((5, 5, 16, 16), sig_dims=2)
        >>> s.flatten_sig()
        (5, 5, 256)
        """
        return Shape(shape=self._nav_shape + (self.sig.size,), sig_dims=1)

    @property
    def dims(self) -> int:
        """
        Number of dimensions

        Examples
        --------

        >>> from libertem.common import Shape
        >>> s = Shape((5, 5, 16, 16), sig_dims=2)
        >>> s.dims
        4
        >>> s.nav.dims  # creates a new temporary Shape and accesses .dims on it
        2
        >>> s.sig.dims
        2
        """
        return self.nav_dims + self.sig_dims

    @property
    def nav_dims(self) -> int:
        return len(self._nav_shape)

    @property
    def sig_dims(self) -> int:
        return len(self._sig_shape)

    def __iter__(self) -> Iterator[int]:
        """
        Iterate over all parts of the shape
        """
        return iter(self._nav_shape + self._sig_shape)

    def __repr__(self) -> str:
        return repr(tuple(self))

    @overload
    def __getitem__(self, k: int) -> int: ...

    @overload
    def __getitem__(self, k: slice) -> tuple[int, ...]: ...

    def __getitem__(self, k):
        return tuple(self)[k]

    def __hash__(self):
        return hash(((self._nav_shape + self._sig_shape), self._sig_dims))

    def __len__(self) -> int:
        return len(self._sig_shape) + len(self._nav_shape)

    def __eq__(self, other: object) -> bool:
        """
        Shape instances are equal if both the shape tuple and the number of signal dimensions
        are equal.
        """
        if not isinstance(other, Shape):
            raise NotImplementedError()
        dims_eq = self._sig_dims == other._sig_dims
        values_eq = tuple(self) == tuple(other)
        return dims_eq and values_eq

    def __add__(self, other: object) -> "Shape":
        """
        Right addition of a Shape object and a tuple.
        Right addition adds the tuple to the signal dimensions of the Shape object.
        """
        if isinstance(other, tuple):
            return Shape(self._nav_shape + self._sig_shape + other,
                 sig_dims=self.sig.dims + len(other))
        else:
            raise NotImplementedError()

    def __radd__(self, other: object) -> "Shape":
        """
        Left addition of a Shape object and a tuple
        Left addition adds the tuple to the navigation dimensions of the Shape object.
        """
        if isinstance(other, tuple):
            return Shape(self._nav_shape + other + self._sig_shape,
                 sig_dims=self.sig.dims)
        else:
            return NotImplemented

    def __getstate__(self) -> dict[str, Any]:
        return {
            k: getattr(self, k)
            for k in self.__slots__
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        for k, v in state.items():
            setattr(self, k, v)


class SigOnlyShape(Shape):
    def __init__(self, shape: "ShapeLike"):
        self._sig_shape = tuple(shape)
        self._nav_shape = tuple()
        self._sig_dims = len(self._sig_shape)

    def __iter__(self):
        return iter(self._sig_shape)

    def __getitem__(self, k):
        return self._sig_shape[k]

    def __len__(self) -> int:
        return len(self._sig_shape)

    def to_tuple(self) -> tuple[int, ...]:
        return self._sig_shape

    @property
    def nav_dims(self) -> int:
        return 0

    @property
    def sig_dims(self) -> int:
        return len(self._sig_shape)

    @property
    def dims(self) -> int:
        return self.sig_dims

    def flatten_nav(self) -> "Shape":
        return self


class NavOnlyShape(Shape):
    def __init__(self, shape: "ShapeLike"):
        self._sig_shape = tuple()
        self._nav_shape = tuple(shape)
        self._sig_dims = 0

    def __iter__(self):
        return iter(self._nav_shape)

    def __getitem__(self, k):
        return self._nav_shape[k]

    def __len__(self) -> int:
        return len(self._nav_shape)

    def to_tuple(self) -> tuple[int, ...]:
        return self._nav_shape

    @property
    def nav_dims(self) -> int:
        return len(self._nav_shape)

    @property
    def sig_dims(self) -> int:
        return 0

    @property
    def dims(self) -> int:
        return self.nav_dims

    def flatten_sig(self) -> "Shape":
        return self


ShapeLike = Union[Shape, Sequence[int]]
